import click
import git
import time
import json
import tempfile
import traceback
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.flowcli import help_option
from convisoappsec.flow import GitAdapter
from convisoappsec.flowcli.common import on_http_error
from convisoappsec.logger import LOGGER

class SameCommitException(Exception):
    pass


@click.command()
@help_option
@click.option(
    "-c",
    "--current-commit",
    required=False,
    help="If no value is given the HEAD commit of branch is used.",
)
@click.option(
    "-p",
    "--previous-commit",
    required=False,
    help="""If no value is given, the value is retrieved from the lastest
    deploy at flow application.""",
)
@click.option(
    "-r",
    "--repository-dir",
    required=False,
    type=click.Path(exists=True, resolve_path=True),
    default='.',
    show_default=True,
    help="Repository directory.",
)
@click.option(
    "--asset-id",
    required=False,
    envvar=("CONVISO_ASSET_ID", "FLOW_ASSET_ID"),
    help="Asset ID on Conviso Platform",
)
@click.pass_context
def values(context, repository_dir, current_commit, previous_commit, asset_id):
    try:
        if context.params['asset_id'] is not None and asset_id is None:
            asset_id = context.params['asset_id']

        git_adapter = GitAdapter(repository_dir)
        commits = deploys_from_asset(asset_id=asset_id)
        current_commit = current_commit or git_adapter.head_commit
        last_commit = commits[0]['currentCommit'] if commits else git_adapter.empty_repository_tree_commit

        if last_commit == current_commit:
            raise SameCommitException(
                "Previous commit ({0}) and Current commit ({1}) are the same, nothing to do."
                .format(last_commit, current_commit)
            )

        if not previous_commit:
            previous_commit = commits[0]['currentCommit'] if commits else git_adapter.empty_repository_tree_commit

            if previous_commit != '4b825dc642cb6eb9a060e54bf8d69288fbee4904':
                try:
                    git_adapter._repo.commit(previous_commit)
                except (git.exc.BadName, ValueError):
                    commits = deploys_from_asset(asset_id=asset_id)
                    previous_commit = None

                    for commit in commits:
                        commit_hash = commit['currentCommit']
                        try:
                            git_adapter._repo.commit(commit_hash)
                            previous_commit = commit_hash
                            break
                        except (git.exc.BadName, ValueError):
                            continue

                    if previous_commit is None:
                        previous_commit = '4b825dc642cb6eb9a060e54bf8d69288fbee4904'

        commit_history_list = git_adapter.get_commit_history()

        commit_history_file = tempfile.NamedTemporaryFile(mode='w+', delete=False)
        json.dump(commit_history_list, commit_history_file)
        commit_history_file.seek(0)

        return {
            "previous_commit": previous_commit,
            "current_commit": current_commit,
            "commit_history": commit_history_file.name,
        }

    except SameCommitException as e:
        LOGGER.warning(str(e))

        return None

    except git.exc.GitError as e:
        LOGGER.error(f"Git error: {str(e)}")
        raise click.ClickException(f"Git error: {str(e)}") from e

    except Exception as e:
        if "GraphqlController" in str(e):
            LOGGER.error(f"GraphQL API error: {str(e)}")
            raise click.ClickException(f"Error communicating with Conviso API: {str(e)}") from e

        LOGGER.warning("There was an error. Our team has been notified.")
        full_trace = traceback.format_exc()
        LOGGER.error(full_trace)
        on_http_error(e)
        raise click.ClickException(str(e)) from e

@pass_flow_context
def deploys_from_asset(flow_context, asset_id, max_retries=3):
    """ Returns all deploys from an asset with retry logic """
    for attempt in range(max_retries):
        try:
            conviso_api = flow_context.create_conviso_graphql_client()
            deploys = conviso_api.deploys.get_deploys_by_asset(asset_id=asset_id)

            return deploys
        except Exception as e:
            if attempt == max_retries - 1:
                LOGGER.error(f"Failed to fetch deploys for asset {asset_id} after {max_retries} attempts: {str(e)}")
                raise
            LOGGER.warning(f"Attempt {attempt + 1} failed: {str(e)}. Retrying...")
            time.sleep(2 ** attempt)
    return None



import tarfile


class SourceCodeCompressor(object):
    TAR_WRITE_MODE = 'w|'
    TAR_ROOT_DIR = '.'

    def __init__(self, source_code_dir="."):
        self.source_code_dir = source_code_dir

    def write_to(self, fileobj):
        tarball_filehandler = tarfile.open(
            mode=self.TAR_WRITE_MODE,
            fileobj=fileobj
        )

        tarball_filehandler.add(
            name=self.source_code_dir,
            arcname=self.TAR_ROOT_DIR,
        )

        tarball_filehandler.close()



import click

from convisoappsec.flowcli import help_option

from .time_ import time_
from .versioning_style import versioning_style


@click.group()
@help_option
def sort_by():
    pass


sort_by.add_command(time_)
sort_by.add_command(versioning_style)

sort_by.epilog = '''
  Run flow deploy create with tag-tracker sort-by COMMAND --help
  for more information on a command.
'''



from convisoappsec.flow.graphql_api.beta.models.issues.sast import CreateSastFindingInput

class CreateIacFindingInput(CreateSastFindingInput):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import requests
import jmespath

from convisoappsec.common.graphql.errors import AuthenticationError, Error, ResponseError, ServerError


class RequestErrorHandler:
    def __init__(self, error):
        self.error = error

    def handle_request_error(self):
        try:
            raise self.error
        except requests.exceptions.HTTPError as e:
            self._handle_http_error(e)
        except Exception as e:
            self._handle_unexpected_error(e)

    @staticmethod
    def _handle_unexpected_error(error):
        raise Error(str(error)) from error

    @staticmethod
    def _handle_http_error(error):
        if error.response.status_code == 401:
            error_msg_fmt = 'Unauthorized access, check your credentials. {}'
            response = error.response.json()
            error_messages = response.get('errors', [])
            error_msg = error_msg_fmt.format(error_messages)

            raise AuthenticationError(error_msg) from error

        if error.response.status_code == 500:
            error_msg_fmt = 'Internal Server Error. {}'
            error_msg = error_msg_fmt.format(error.response.text)

            raise ServerError(error_msg) from error

        raise Error(error.response.text) from error


class GraphQlErrorHandler:
    def __init__(self, response):
        self.response = response

    def raise_on_graphql_body_error(self):
        data = self.response.get('data', [])

        for key in data:
            if data[key] is None:
                continue

            errors = data[key].get('errors', [])
            has_errors = len(errors) > 0
            if has_errors:
                raise ResponseError(errors)

    def raise_on_graphql_error(self):
        errors = self.response.get('errors', [])

        if not errors:
            return

        error = errors[0]

        error_path = 'extensions.code'
        
        code = jmespath.search(
            error_path,
            error
        )

        message = error.get('message', '')

        raise ResponseError(message, code)



from enum import Enum
from functools import reduce


class CIProvider(Enum):
    AWS_CODEBUILD = {
        '_env_vars': [
            'CODEBUILD_BUILD_ARN',
            'CODEBUILD_BUILD_ID',
            'CODEBUILD_BUILD_NUMBER'
        ]
    }
    AZURE_PIPELINES = {
        '_env_vars': [
            'SYSTEM_JOBDISPLAYNAME',
            'SYSTEM_JOBID',
            'SYSTEM_TEAMPROJECT'
        ]
    }
    BITBUCKET = {
        '_env_vars': [
            'BITBUCKET_PROJECT_KEY',
            'BITBUCKET_PROJECT_UUID',
            'BITBUCKET_PIPELINE_UUID'
        ]
    }
    CIRCLECI = {
        '_env_vars': [
            'CIRCLECI',
            'CIRCLE_JOB',
            'CIRCLE_USERNAME'
        ]
    }
    CODEFRESH = {
        '_env_vars': [
            'CF_REPO_NAME',
            'CF_REVISION',
            'CF_BRANCH'
        ]
    }
    GITLAB = {
        '_env_vars': [
            'GITLAB_CI',
            'CI_PROJECT_ID',
            'CI_SERVER_NAME'
        ]
    }
    GITHUB = {
        '_env_vars': [
            'GITHUB_REPOSITORY',
            'GITHUB_REF',
            'GITHUB_JOB'
        ]
    }
    JENKINS = {
        '_env_vars': [
            'BUILD_NUMBER',
            'BUILD_TAG',
            'JOB_NAME'
        ]
    }
    OTHER = {
        '_env_vars': []
    }


    @classmethod
    def names(cls):
        return [provider.name for provider in cls]


    def env_vars_exists(self, env):
        provider_vars = self.__provider_vars

        if not provider_vars:
            return False

        compute_found_env_vars = ComputeFoundEnvVars(env)

        found_vars = reduce(compute_found_env_vars, provider_vars, 0)
        vars_length = len(provider_vars)

        return found_vars == vars_length


    @property
    def __provider_vars(self):
        return self.value['_env_vars']


class ComputeFoundEnvVars:
    def __init__(self, environment):
        self.__environment = environment

    def __call__(self, found_vars, env_var_name):
        if self.__environment.get(env_var_name):
            return found_vars + 1
        else:
            0


import click
import click_log

from convisoappsec.common.git_data_parser import GitDataParser
from convisoappsec.flow.graphql_api.v1.models.asset import AssetInput
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import on_http_error
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.logger import LOGGER

click_log.basic_config(LOGGER)


def parse_repository_name(repository_dir, asset_name=None):
    has_user_input = asset_name != None
    if has_user_input:
        return asset_name

    return GitDataParser(repository_dir).parse_name()


@click.command()
@click_log.simple_verbosity_option(LOGGER)
@click.option(
    "-c",
    "--company-id",
    type=int,
    required=True,
    help="The Company ID from your organization in Conviso Platform.",
)
@click.option(
    "-r",
    "--repository-dir",
    type=click.Path(exists=True, resolve_path=True),
    default=".",
    show_default=True,
    required=True,
    help="The directory path for the asset.",
)
@click.option(
    "--name",
    type=str,
    default=None,
    show_default=False,
    required=False,
    help="Customize the Asset name.",
)
@click.option(
    "--scan-type",
    type=str,
    default="None",
    show_default=True,
    required=False,
    help="Customize the Asset scan type.",
)
@help_option
@pass_flow_context
def create(flow_context, company_id, repository_dir, name, scan_type):
    try:
        conviso_api = flow_context.create_conviso_graphql_client()

        perform_command(conviso_api, company_id, repository_dir, name, scan_type)

    except Exception as exception:
        on_http_error(exception)
        raise click.ClickException(str(exception)) from exception


def perform_command(
    conviso_api,
    company_id,
    repository_dir,
    name,
    scan_type,
):
    asset_name = parse_repository_name(repository_dir, name)

    asset_model = AssetInput(
        company_id,
        asset_name,
        scan_type,
    )

    print('Creating new asset to repository: "{}"'.format(repository_dir))
    asset = conviso_api.assets.create_asset(asset_model)

    asset_url = conviso_api.assets.get_asset_url(company_id, asset["id"])
    print("The created Asset is available at Conviso Platform: {}".format(asset_url))



import click

from convisoappsec.flowcli import help_option
from .ls import ls
from .create import create


@click.group()
@help_option
def assets():
    pass


assets.add_command(ls)
assets.add_command(create)


assets.epilog = '''
  Run 'conviso assets COMMAND --help' for more information on a command.
'''



import os

import click
import click_log
from convisoappsec.flow import api
from convisoappsec.flow.util.ci_provider import CIProvider
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import process_ci_provider_option
from convisoappsec.flowcli.iac.entrypoint import iac
from convisoappsec.logger import LOGGER
from convisoappsec.version import __version__

from .ast import ast
from .context import pass_flow_context
from .findings import findings
from .sast import sast
from .sca import sca
from .vulnerability import vulnerability
from .assets import assets
from .sbom import sbom
from .container import container

click_log.basic_config(LOGGER)

@click.group()
@click_log.simple_verbosity_option(LOGGER, '-l', '--verbosity')
@click.option(
    '-k',
    '--api-key',
    show_envvar=True,
    envvar=("CONVISO_API_KEY", "FLOW_API_KEY"),
    help="The api key to access Conviso Platform resources.",
)
@click.option(
    '-u',
    '--api-url',
    show_envvar=True,
    envvar=("CONVISO_API_URL", "FLOW_API_URL"),
    default=api.DEFAULT_API_URL,
    show_default=True,
    help='The api url to access Conviso Platform resources.',
)
@click.option(
    '-i',
    '--api-insecure',
    show_envvar=True,
    envvar=("CONVISO_API_INSECURE", "FLOW_API_INSECURE"),
    default=False,
    show_default=True,
    is_flag=True,
    help='HTTPS requests to untrusted hosts is enable.',
)
@click.option(
    '-c',
    '--ci-provider-name',
    show_envvar=True,
    envvar="CI_PROVIDER_NAME",
    type=click.Choice(CIProvider.names()),
    default=None,
    show_default=True,
    required=False,
    help="The ci provider used by project. "
         "When not informed, an automatic search will be performed."
)
@click.option(
    '--experimental',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable experimental features.",
)
@help_option
@click.version_option(
    __version__,
    '-v',
    '--version',
    message='%(prog)s %(version)s'
)
@pass_flow_context
def cli(flow_context, api_key, api_url, api_insecure, experimental, ci_provider_name):
    flow_context.key = api_key
    flow_context.url = api_url
    flow_context.insecure = api_insecure
    flow_context.experimental = experimental

    ci_provider = process_ci_provider_option(ci_provider_name, os.environ)
    flow_context.ci_provider = ci_provider
    LOGGER.debug('CI provider name detected: {}'.format(flow_context.ci_provider.name))


cli.add_command(findings)
cli.add_command(sast)
cli.add_command(sca)
cli.add_command(vulnerability)
cli.add_command(ast)
cli.add_command(iac)
cli.add_command(assets)
cli.add_command(sbom)
cli.add_command(container)

cli.epilog = '''
  Run conviso COMMAND --help for more information on a command.
'''



import click
from convisoappsec.flowcli.common import on_http_error
from convisoappsec.common import safe_join_url
from convisoappsec.flow.graphql_api.v1.client import ConvisoGraphQLClient

class Companies():
    def ls(self, flow_context, company_id=None):
        api_key = flow_context.key

        try:
            url = safe_join_url(flow_context.url, "/graphql")
            conviso_api = ConvisoGraphQLClient(api_url=url,api_key=api_key)

            return perform_command(conviso_api, company_id)
        except Exception as exception:
            on_http_error(exception)
            raise click.ClickException(str(exception)) from exception

def perform_command(conviso_api, company_id):
    if company_id is not None:
        companies = conviso_api.companies.get_company_by_id(company_id)
    else:
        companies = conviso_api.companies.get_companies()

    return companies



from convisoappsec.flow.graphql_api.beta.models.issues.normalize import Normalize

class CreateSastFindingInput:
    def __init__(
        self,
        asset_id,
        code_snippet,
        file_name,
        vulnerable_line,
        first_line,
        title,
        description,
        severity,
        reference,
        category,
        original_issue_id_from_tool,
        solution,
        control_sync_status_id
    ):
        self.asset_id = asset_id
        self.severity = Normalize.normalize_severity(severity)
        self.title = title
        self.description = description
        self.code_snippet = code_snippet
        self.file_name = file_name
        self.vulnerable_line = int(vulnerable_line)
        self.first_line = int(first_line)
        self.reference = reference
        self.category = category
        self.original_issue_id_from_tool = original_issue_id_from_tool
        self.solution = solution
        self.control_sync_status_id = control_sync_status_id

    def to_graphql_dict(self):
        """
        This function returns a dictionary containing various attributes of an
        asset in a GraphQL format.
        """
        return {
            "assetId": int(self.asset_id),
            "severity": self.severity,
            "title": self.title,
            "description": self.description,
            "codeSnippet": self.code_snippet,
            "fileName": self.file_name,
            "vulnerableLine": int(self.vulnerable_line),
            "firstLine": int(self.first_line),
            "reference": self.reference,
            "category": str(self.category),
            "originalIssueIdFromTool": str(self.original_issue_id_from_tool),
            "solution": str(self.solution),
            "controlSyncStatusId": str(self.control_sync_status_id)
        }



import json
import tempfile
import re
import yaml
import git
import os
from contextlib import suppress
from convisoappsec.logger import LOGGER
from git.exc import GitCommandError


class GitAdapter(object):
    LIST_OPTION = '--list'
    SORT_OPTION = '--sort'
    FORMAT_OPTION = '--format'
    ANCESTRY_PATH_OPTION = '--ancestry-path'
    HEAD_COMMIT = 'HEAD'
    OPTION_WITH_ARG_FMT = '{option}={value}'
    EMPTY_REPOSITORY_HASH = '4b825dc642cb6eb9a060e54bf8d69288fbee4904'

    def __init__(self, repository_dir='.', load_remote_repositories_heads=False, unshallow_repository=False):
        LOGGER.debug('Unshallow: {}'.format(unshallow_repository))
        LOGGER.debug('Load remote: {}'.format(load_remote_repositories_heads))

        self._git_client = git.cmd.Git(repository_dir)
        self._first_commit = None
        self._repo = git.Repo(repository_dir)

        if load_remote_repositories_heads:
            self.__load_remote_repositories_heads()

        if unshallow_repository:
            self.__unshallow_repository()

    def repo_url(self):
        """
        Function to get the repository URL and convert it to an HTTPS format if necessary.

        Returns:
            str: The repository URL in HTTPS format, or None if the URL cannot be determined.
        """
        try:
            repos_url = self._repo.remotes.origin.url
        except AttributeError:
            if self._repo.remotes:
                repos_url = self._repo.remotes[0].url
            else:
                return None

        if repos_url.startswith('git@'):
            return repos_url.replace(':', '/').replace('git@', 'https://').replace('.git', '')

        elif repos_url.startswith('ssh://git@ssh.dev.azure.com'):
            parts = repos_url.split('/')
            if len(parts) >= 7:
                organization = parts[4]
                project = parts[5]
                repo = parts[6].replace('.git', '')
                return f"https://dev.azure.com/{organization}/{project}/_git/{repo}"

        return repos_url

    def get_branch_name(self):
        """retrieves the branch name"""
        try:
            return self._repo.active_branch.name
        except TypeError:
            return 'HEAD'


    def get_commit_history(self):
        """
        Retrieves the commit history (including stashes) of the repository.

        Returns:
            list: A list of commit information.
        """
        commits = self._repo.iter_commits()

        commit_info_list = [
            {
                'commit': commit.hexsha,
                'author': "{author_name} <{author_email}>".format(
                    author_name=commit.author.name, author_email=commit.author.email
                ),
                'date': commit.authored_datetime.strftime("%Y-%m-%d %H:%M:%S"),
                'message': commit.message
            }
            for commit in commits
        ]

        # Include information about stashes
        stashes = self._repo.git.stash("list", "--format=%H|%gd|%ci|%P|%gs", "--date=iso").splitlines()
        stash_info_list = [
            {
                'commit': stash_info.split('|')[0],
                'stash_ref': stash_info.split('|')[1],
                'date': stash_info.split('|')[2],
                'parent_commits': stash_info.split('|')[3].split(),
                'message': stash_info.split('|')[4],
                'contributors': self.get_contributors_for_stash_merge(stash_info.split('|')[0])
            }
            for stash_info in stashes
        ]

        commit_info_list.extend(stash_info_list)

        return commit_info_list

    def get_contributors_for_stash_merge(self, stash_commit):
        """
        Get contributors involved in a stash merge.

        Args:
            stash_commit (str): The commit hash of the stash.

        Returns:
            list: List of contributors involved in the stash merge.
        """
        contributors = set()
        stash_diff = self._repo.git.diff(stash_commit + "^", stash_commit, "--name-only").splitlines()

        for file_path in stash_diff:
            try:
                blame_output = self._repo.git.blame(stash_commit, "--", file_path, p=True).splitlines()
                author_info = {}
                for line in blame_output:
                    if line.startswith('author '):
                        author_info['name'] = line[len('author '):]
                    elif line.startswith('author-mail '):
                        author_info['email'] = line[len('author-mail '):].strip('<>')

                    if 'name' in author_info and 'email' in author_info:
                        contributors.add("{name} {email}".format(name=author_info['name'], email=author_info['email']))
                        author_info = {}

            except GitCommandError as e:
                LOGGER.warning(f"Could not process git blame for {file_path}: {e}")


        return list(contributors)

    def tags(self, sort='-committerdate'):
        sort_option = self.OPTION_WITH_ARG_FMT.format(
            option=self.SORT_OPTION,
            value=sort,
        )

        args = (self.LIST_OPTION, sort_option)
        client_output = self._git_client.tag(args)
        tags = client_output.splitlines()
        return tags

    def diff(self, version, another_version):
        version = version or self.EMPTY_REPOSITORY_HASH

        if version == self.EMPTY_REPOSITORY_HASH:
            msg_fmt = """Creating diff comparing revision[{0}] and the repository beginning"""
            LOGGER.warning(msg_fmt.format(another_version))

        diff_file = tempfile.TemporaryFile()
        self._git_client.diff(version, another_version, output_stream=diff_file)

        return diff_file

    def diff_stats(self, version, another_version):
        version = version or self.EMPTY_REPOSITORY_HASH

        if version == self.EMPTY_REPOSITORY_HASH:
            msg_fmt = """Creating diff stats comparing revision[{0}] and the repository beginning"""
            LOGGER.warning(msg_fmt.format(another_version))

        stats_output = tempfile.TemporaryFile()
        self._git_client.diff(version, another_version, '--numstat', output_stream=stats_output)

        stats_summary = GitDiffNumStatSummary.load(stats_output)

        return stats_summary

    @property
    def first_commit(self):
        if self._first_commit:
            return self._first_commit

        command_output = tempfile.TemporaryFile()

        args = [
            '--reverse',
            "--pretty=%H",
        ]

        self._git_client.log(args, output_stream=command_output)
        command_output.seek(0)
        first_in_bytes = command_output.readline()
        command_output.close()
        first = first_in_bytes.decode()

        return first.strip()

    def commit_is_first(self, commit):
        return commit == self.first_commit

    @property
    def head_commit(self):
        client_output = self._git_client.rev_parse(self.HEAD_COMMIT)
        return client_output.strip()

    @property
    def current_commit(self):
        return self.head_commit

    @property
    def previous_commit(self):
        return self.previous_commit_from(self.current_commit)

    def previous_commit_from(self, commit, offset=1):
        if self.commit_is_first(commit):
            return self.EMPTY_REPOSITORY_HASH

        command_fmt = "{commit}~{offset}"

        command = command_fmt.format(
            commit=commit,
            offset=offset,
        )

        client_output = self._git_client.rev_parse(
            command
        )

        return client_output.strip()

    def show_commit_refs(self, commit):
        with tempfile.TemporaryFile() as client_output:
            self._git_client.show_ref(
                "--head", "--heads", "--tags", output_stream=client_output
            )
            refs = _read_file_lines_generator(client_output)
            refs = list(
                filter(
                    lambda ref: re.search(commit, ref),
                    refs,
                )
            )

            return refs

    def show_commit_from_tag(self, tag):
        client_output = self._git_client.rev_parse(
            tag
        )

        return client_output.strip()

    def get_commit_author(self, commit_hash):
        if self.EMPTY_REPOSITORY_HASH == commit_hash:
            default_commit = {
                'name': 'Default',
                'email': 'Default',
                'commit': commit_hash
            }
            return default_commit

        delimiter = '|;|'
        fmt_author_name = '%an'
        fmt_author_email = '%ae'
        fmt_long_commit_hash = '%H'

        row_fmt = '{name}{delimiter}{email}{delimiter}{commit}'.format(
            name=fmt_author_name,
            delimiter=delimiter,
            email=fmt_author_email,
            commit=fmt_long_commit_hash
        )
        format_option = self.OPTION_WITH_ARG_FMT.format(
            option=self.FORMAT_OPTION,
            value=row_fmt,
        )

        author = self._git_client.show(
            '-s', format_option, commit_hash
        ).split(delimiter)

        author_data = {
            'name': author[0],
            'email': author[1],
            'commit': author[2],
        }

        return author_data

    def get_commits_by_range(self, start_commit, end_commit):
        tmp_commits = tempfile.TemporaryFile()

        self._git_client.rev_list(
            self.ANCESTRY_PATH_OPTION, start_commit + '..' + end_commit, output_stream=tmp_commits
        )

        return _read_file_lines_generator(tmp_commits)

    def get_commit_authors_by_range(self, start_commit, end_commit):
        start_commit_range = start_commit
        authors = []

        if self.EMPTY_REPOSITORY_HASH == start_commit_range:
            start_commit_range = self.first_commit

        commit_range = f"{start_commit_range}..{end_commit}"
        log_output = self._git_client.log(commit_range, "--pretty=tformat:%H|%an|%ae").splitlines()

        for line in log_output:
            commit_hash, name, email = line.split('|')
            author_data = {
                'name': name,
                'email': email,
                'commit': commit_hash,
            }
            authors.append(author_data)

        return authors

    @property
    def empty_repository_tree_commit(self):
        return self.EMPTY_REPOSITORY_HASH

    @property
    def remote_repositories_name(self):
        args = ("show")
        client_output = self._git_client.remote(args)
        repositories = client_output.splitlines()
        return repositories

    def __load_remote_repositories_heads(self):
        heads_refspec_format = "refs/heads/*:refs/remotes/{remote_repository_name}/*"

        for remote_repository_name in self.remote_repositories_name:
            try:
                heads_refspec = heads_refspec_format.format(
                    remote_repository_name=remote_repository_name
                )

                args = (remote_repository_name, heads_refspec)
                self._git_client.fetch(args)

            except GitCommandError:
                raw_msg = "We can\'t ensure that the refspec refs/heads/* from repository {repository} were loaded."
                msg = raw_msg.format(repository=remote_repository_name)
                LOGGER.warning(msg)

    @property
    def is_shallow_repository(self):
        import os.path

        args = ('--git-dir')
        git_dir = self._git_client.rev_parse(args)

        working_dir = self._git_client.working_dir
        shallow_file = os.path.join(working_dir, git_dir, 'shallow')

        return os.path.isfile(shallow_file)

    def __unshallow_repository(self):
        if not self.is_shallow_repository:
            return

        args = ('--unshallow')
        self._git_client.fetch(args)


def _read_file_lines_generator(file):
    file.seek(0)

    while True:
        line = file.readline()
        line = line.decode().strip()

        if line:
            yield line
        else:
            break


class InvalidGitDiffNumStatLineValueException(ValueError):
    pass


class GitDiffNumStatLine(object):
    ADDED_LINES_POSITION = 1
    DELETED_LINES_POSITION = 2
    FILE_PATH_POSITION = 3

    # (added_lines) (deleted_lines) (file_path)
    SRC_LINE_REGEX = r'(\d+)\s+(\d+)\s+(.*)'
    BIN_LINE_REGEX = r'(-)\s+(-)\s+(.*)'

    def __init__(self, added_lines, deleted_lines, file_path):
        self.added_lines = added_lines
        self.deleted_lines = deleted_lines
        self.file_path = file_path

    @classmethod
    def parse(cls, raw_line):
        with suppress(AttributeError):
            match = re.match(cls.SRC_LINE_REGEX, raw_line)
            group = match.group

            added_lines_str = group(cls.ADDED_LINES_POSITION)
            deleted_lines_str = group(cls.DELETED_LINES_POSITION)
            file_path = group(cls.FILE_PATH_POSITION)

            added_lines = int(added_lines_str)
            deleted_lines = int(deleted_lines_str)

            return cls(added_lines, deleted_lines, file_path)

        with suppress(AttributeError):
            match = re.match(cls.BIN_LINE_REGEX, raw_line)

            file_path = match.group(cls.FILE_PATH_POSITION)

            return cls(0, 0, file_path)

        error_msg_fmt = '\n'.join([
            'The expected git diff numstat line format are:',
            'Expected format: {src_fmt}',
            'Expected format: {bin_fmt}',
            'Given value: {given}',
        ])

        msg = error_msg_fmt.format(
            src_fmt=cls.SRC_LINE_REGEX,
            bin_fmt=cls.BIN_LINE_REGEX,
            given=raw_line
        )

        raise InvalidGitDiffNumStatLineValueException(msg)

    @classmethod
    def load(cls, numstat_fh):
        numstat_fh.seek(0)

        while True:
            line = numstat_fh.readline()

            with suppress(AttributeError):
                line = line.decode()

            if line:
                yield cls.parse(line)
                continue

            break


class GitDiffNumStatSummary(object):

    def __init__(self):
        self.added_lines = 0
        self.deleted_lines = 0
        self.changed_files = []

    def _add_numstat_lines(self, numstat_lines):
        for numstat_line in numstat_lines:
            self._add_numstat_line(numstat_line)

    def _add_numstat_line(self, numstat_line):
        self._add_added_lines(numstat_line.added_lines)
        self._add_deleted_lines(numstat_line.deleted_lines)
        self._add_changed_files(numstat_line.file_path)

    def _add_added_lines(self, added_lines):
        self.added_lines += added_lines

    def _add_deleted_lines(self, deleted_lines):
        self.deleted_lines += deleted_lines

    def _add_changed_files(self, changed_file):
        self.changed_files.append(
            changed_file
        )

    @property
    def changed_lines(self):
        return self.added_lines + self.deleted_lines

    @classmethod
    def load(cls, numstat_fh):
        git_diffnumstat_lines = GitDiffNumStatLine.load(numstat_fh)
        summary = cls()
        summary._add_numstat_lines(git_diffnumstat_lines)

        return summary

    @property
    def dict(self):
        field_names = [
            'added_lines',
            'deleted_lines',
            'changed_lines',
            'changed_files',
        ]

        fields = {
            name: getattr(self, name) for name in field_names
        }

        return fields


def quoted_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='"')


yaml.add_representer(str, quoted_presenter)


class CommitAuthorFile:
    def __init__(self):
        self._tmp_authors = tempfile.NamedTemporaryFile()

    def add_author(self, author):
        yaml_str = yaml.dump(author, explicit_start=True)
        yaml_bytes = yaml_str.encode()
        self._tmp_authors.write(yaml_bytes)

    def get_file_descriptor(self):
        self._tmp_authors.seek(0)
        return self._tmp_authors



import io
import sys
from shlex import quote

import click
import requests

from convisoappsec.flow.util.ci_provider import CIProvider

class CreateDeployException(Exception):
    pass

class PerformDeployException(Exception):
    pass

class DeployFormatter(object):
    DEFAULT = 'default'
    ENV_VARS = 'env_vars'

    def __init__(self, format):
        self.stategy = self._select_strategy(format)

    def format(self, deploy):
        return self.stategy(deploy)

    def _select_strategy(self, format):
        format_strategies = {
            self.DEFAULT: self.default_format,
            self. ENV_VARS: self.env_vars_format,
        }

        strategy = format_strategies.get(format)

        if not strategy:
            msg_fmt = "Allowed deploy formats[{0}]. Given {1}"
            msg = msg_fmt.format(
                "|".join(self.FORMATS()),
                format,
            )

            raise ValueError(msg)

        return strategy

    @classmethod
    def FORMATS(cls):
        return [
            cls.DEFAULT,
            cls.ENV_VARS,
        ]

    def env_vars_format(self, deploy):
        to_env_vars_translation_args = [
            # (env_var_name, attrib_name, attrib_default_val)
            ('FLOW_DEPLOY_ID', 'id', 0),
            ('FLOW_DEPLOY_CREATED_AT', 'created_at', ''),
            ('FLOW_DEPLOY_CURRENT_VERSION_TAG', 'current_tag', ''),
            ('FLOW_DEPLOY_PREVIOUS_VERSION_TAG', 'previous_tag', ''),
            ('FLOW_DEPLOY_CURRENT_VERSION_COMMIT', 'current_commit', ''),
            ('FLOW_DEPLOY_PREVIOUS_VERSION_COMMIT', 'previous_commit', ''),
            ('CONVISO_DEPLOY_ID', 'id', 0),
            ('CONVISO_DEPLOY_CREATED_AT', 'created_at', ''),
            ('CONVISO_DEPLOY_CURRENT_VERSION_TAG', 'current_tag', ''),
            ('CONVISO_DEPLOY_PREVIOUS_VERSION_TAG', 'previous_tag', ''),
            ('CONVISO_DEPLOY_CURRENT_VERSION_COMMIT', 'current_commit', ''),
            ('CONVISO_DEPLOY_PREVIOUS_VERSION_COMMIT', 'previous_commit', ''),
        ]

        env_vars = {}

        for arg in to_env_vars_translation_args:
            env_var_name, attrib_name, attrib_default_val = arg

            env_var_val = env_vars[env_var_name] = deploy.get(
                attrib_name, attrib_default_val
            )

            env_vars[env_var_name] = str(env_var_val)

        buffer = io.StringIO()

        for var_name, var_val in env_vars.items():
            line_fmt = "export {name}={value}"
            line = line_fmt.format(
                name=quote(var_name),
                value=quote(var_val),
            )

            print(line, file=buffer)

        buffer.seek(0)
        return buffer.read()

    def default_format(self, src_deploy):
        deploy = {
            'id': 0,
            'current_tag': '',
            'previous_tag': '',
            'current_commit': '',
            'previous_commit': '',
            'created_at': '',
        }

        deploy.update(src_deploy)

        default_fmt = '''
Deploy stats:
  current_version.commit={current_commit}
  current_version.tag={current_tag}
  previous_version.commit={previous_commit}
  previous_version.tag={previous_tag}
        '''

        return default_fmt.format(
            **deploy
        )


def notify_created_deploy(deploy):
    formatter = DeployFormatter(DeployFormatter.DEFAULT)
    click.echo(formatter.format(deploy))


def asset_id_option(*args, **kwargs):
    kwargs["envvar"] = kwargs.get("envvar", ("CONVISO_ASSET_ID", "FLOW_ASSET_ID"))
    kwargs["show_envvar"] = kwargs.get("show_envvar", True)
    kwargs["required"] = kwargs.get("required", False)
    kwargs["hidden"] = kwargs.get("hidden", False)
    asset_id_param = "--asset-id"
    paramsdecl = args

    if asset_id_param not in args:
        paramsdecl = args + (asset_id_param, )

    return click.option(
        *paramsdecl,
        type=int,
        **kwargs
    )

def on_http_error(error):
    try:
        raise error
    except requests.exceptions.HTTPError as e:
        message = error.response.text

        if not message:
            return

        click.echo(
            "HttpErrorResponse: {0}".format(message),
            file=sys.stderr
        )


def __try_retrieve_ci_provider(provider_name):
    try:
        return CIProvider[provider_name]
    except KeyError as e:
        error_msg_fmt = 'Received for parameter provider_name[{}]. Expected values are: [{}]'
        expected_values = '|'.join(CIProvider.names())
        error_msg = error_msg_fmt.format(provider_name, expected_values)

        raise ValueError(error_msg) from e


def process_ci_provider_option(provider_name, env={}):
    if provider_name is not None:
        return __try_retrieve_ci_provider(provider_name)

    for provider in CIProvider:
        if provider.env_vars_exists(env):
            return provider

    return CIProvider.OTHER



import tempfile
import tarfile
import docker
import os
import click
import json
import docker.errors
from io import BytesIO
from contextlib import suppress
from pathlib import Path
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.flow import GitAdapter

bitbucket = os.getenv('BITBUCKET_CLONE_DIR')


class SASTBox(object):
    REGISTRY = 'public.ecr.aws/convisoappsec'
    REPOSITORY_NAME = 'sastbox_v2'
    DEFAULT_TAG = 'unstable'
    CONTAINER_CODE_DIR = bitbucket or '/code'
    CONTAINER_REPORTS_DIR = '/tmp'
    WORKSPACE_REPORT_PATH = CONTAINER_CODE_DIR
    JSON_REPORT_PATTERN = 'output.json'
    SUCCESS_EXIT_CODE = 1
    USER_ENV_VAR = "USER"

    def __init__(self, registry=None, repository_name=None, tag=None):
        self.docker = docker.from_env(
            version="auto"
        )
        self.container = None
        self.registry = registry or self.REGISTRY
        self.repository_name = repository_name or self.REPOSITORY_NAME
        self.tag = tag or self.DEFAULT_TAG

    def login(self, password, username='AWS'):
        login_args = {
            'registry': self.REGISTRY,
            'username': username,
            'password': password,
            'reauth': True,
        }

        login_result = self.docker.login(**login_args)
        return login_result

    def run_scan_diff(self, code_dir, current_commit, previous_commit, log=None):
        return self._scan_diff(code_dir, current_commit, previous_commit, log)

    @property
    def size(self):
        try:
            registry_data = self.docker.images.get_registry_data(
                self.image
            )
            descriptor = registry_data.attrs.get('Descriptor', {})
            return descriptor.get('size') * 1024 * 1024
        except docker.errors.APIError:
            return 6300 * 1024 * 1024

    def pull(self):
        size = self.size
        layers = {}
        for line in self.docker.api.pull(
                self.repository, tag=self.tag, stream=True, decode=True
        ):
            status = line.get('status', '')
            detail = line.get('progressDetail', {})

            if status == 'Downloading':
                with suppress(Exception):
                    layer_id = line.get('id')
                    layer = layers.get(layer_id, {})
                    layer.update(detail)
                    layers[layer_id] = layer

                    for layer in layers.values():
                        current = layer.get('current')
                        total = layer.get('total')

                        if (current / total) > 0.98 and not layer.get('done'):
                            yield current
                            layer.update({'done': True})

        yield size

    def _scan_diff(self, code_dir, current_commit, previous_commit, log):
        environment = {
            'PREVIOUS_COMMIT': previous_commit,
            'CURRENT_COMMIT': current_commit,
            'SASTBOX_REPORTS_DIR': self.CONTAINER_REPORTS_DIR,
            'SASTBOX_REPORT_DIR': '/tmp',
            'SASTBOX_REPORT_PATTERN': '*.sarif',
            'SASTBOX_CODE_DIR': self.CONTAINER_CODE_DIR,
        }

        command_parts = [
            'ruby', 'manager/sastbox_cli.rb',
            '-c', self.CONTAINER_CODE_DIR,
            '-a',
            '-o', '/tmp/output.sarif',
            f'--diff={previous_commit},{current_commit}',
            '&&',
            'cp', '$(find "$SASTBOX_REPORT_DIR" -name "$SASTBOX_REPORT_PATTERN")', '$SASTBOX_REPORTS_DIR'
        ]
        command = ' '.join(command_parts)

        # Configure container creation
        create_args = {
            'image': self.image,
            'entrypoint': ['sh', '-c'],
            'command': [command],
            'tty': True,
            'detach': True,
            'environment': environment,
        }

        try:
            try:
                self.container = self.docker.containers.create(**create_args)
            except docker.errors.APIError as e:
                raise RuntimeError(f"Failed to create container: {e}")

            # Create and upload source code tarball
            source_code_tarball_file = tempfile.TemporaryFile()
            try:
                source_code_tarball = tarfile.open(mode="w|gz", fileobj=source_code_tarball_file)
                source_code_tarball.add(
                    name=code_dir,
                    arcname=self.CONTAINER_CODE_DIR,
                    filter=lambda tarinfo: tarinfo if not tarinfo.name.endswith('.zip') else None
                )
                source_code_tarball.close()
                source_code_tarball_file.seek(0)
                try:
                    self.container.put_archive("/", source_code_tarball_file)
                except docker.errors.APIError as e:
                    raise RuntimeError(f"Failed to upload tarball: {e}")
            finally:
                source_code_tarball_file.close()

            # Start the container and stream logs
            try:
                self.container.start()
            except docker.errors.APIError as e:
                raise RuntimeError(f"Failed to start container: {e}")

            for line in self.container.logs(stream=True):
                if log:
                    log(line, new_line=False)

            self.recovery_technologies_file()

            wait_result = self.container.wait()
            status_code = wait_result.get('StatusCode')

            if status_code != self.SUCCESS_EXIT_CODE:
                logs = self.container.logs().decode('utf-8')
                raise RuntimeError(f"SASTBox exited with status code {status_code}\nLogs:\n{logs}")

            # Retrieve and extract reports
            try:
                chunks, _ = self.container.get_archive(self.CONTAINER_REPORTS_DIR)
            except docker.errors.APIError as e:
                raise RuntimeError(f"Failed to retrieve reports: {e}")

            reports_tarball_file = tempfile.TemporaryFile()
            try:
                for chunk in chunks:
                    reports_tarball_file.write(chunk)
                tempdir = tempfile.mkdtemp()
                reports_tarball_file.seek(0)
                reports_tarball = tarfile.open(mode="r|", fileobj=reports_tarball_file)
                reports_tarball.extractall(path=tempdir)
                reports_tarball.close()
            finally:
                reports_tarball_file.close()

            # Verify reports exist and return their paths
            reports = self._list_reports_paths(tempdir)
            if not reports:
                raise RuntimeError("No reports found in the container")

        except docker.errors.APIError as e:
            raise RuntimeError(f"Failed to retrieve reports: {e}")

        return reports

    @property
    def repository(self):
        return "{registry}/{repository_name}".format(
            registry=self.registry,
            repository_name=self.repository_name,
        )

    @property
    def image(self):
        return "{repository}:{tag}".format(
            repository=self.repository,
            tag=self.tag,
        )

    def __del__(self):
        with suppress(Exception):
            self.container.remove(v=True, force=True)

    @classmethod
    def _list_reports_paths(cls, root_dir):
        root_dir = root_dir + cls.CONTAINER_REPORTS_DIR
        sastbox_reports_dir = Path(root_dir)

        for report in sastbox_reports_dir.glob(cls.JSON_REPORT_PATTERN):
            yield report

    def recovery_technologies_file(self):
        """ Method to recover a fingerprint file inside the container with founded technology """
        try:
            generator_object, _ = self.container.get_archive('/tmp/')
            file_content = b"".join(generator_object)
            file_content_stream = BytesIO(file_content)
            tar = tarfile.open(fileobj=file_content_stream)
            file_names = tar.getnames()

            fingerprint_file = next(
                (file for file in file_names if file.startswith('tmp/fingerprint') and file.endswith('.json')), None)

            if not fingerprint_file:
                log_func("No file starting with 'fingerprint' and ending with '.json' found.")
                return

            actual_filename = fingerprint_file.split('/')[-1]

            generator_object, _ = self.container.get_archive(f'/tmp/{actual_filename}')
            file_content = b"".join(generator_object)
            file_content_stream = BytesIO(file_content)
            tar = tarfile.open(fileobj=file_content_stream)
            file_data = tar.extractfile(actual_filename)
            content = json.loads(file_data.read())
            technologies = content['result']['technologies']
        except Exception as error:
            msg = "\U0001F4AC Something goes wrong when trying to recover the technologies, continuing ..."
            log_func(msg)
            technologies = []

        if technologies is None:
            return

        self.update_asset_technologies(technologies=technologies)

    @staticmethod
    @pass_flow_context
    @click.pass_context
    def update_asset_technologies(flow_context, context, technologies):
        """
        Update technologies on asset.
        Args:
            flow_context (dict): Flow context containing parameters.
            context (object): Object containing necessary methods (e.g., create_conviso_graphql_client).
            technologies (list): List of technologies to be updated.
        Returns:
            dict: Response from the API call.
        """

        # this prevents a broken execution when something goes wrong.
        try:
            git_adapter = GitAdapter(flow_context.params['repository_dir'])
            repo_url = git_adapter.repo_url()

            company_id = flow_context.params.get('company_id')
            asset_id = flow_context.params.get('asset_id')
            asset_name = flow_context.params.get('asset_name')
            unwanted_technologies = {
                'unknown', 'json', 'text', 'ini', 'diff', 'xml', 'markdown', 'csv', 'gemfile.lock', 'html+erb',
                'javascript+erb', 'robots.txt', 'yaml', 'batchfile', 'java properties', 'svg', 'json with comments'
            }
            updated_technologies = [tech for tech in technologies if tech not in unwanted_technologies]
            conviso_api = context.create_conviso_graphql_client()

            response = conviso_api.assets.update_asset(
                company_id=int(company_id),
                asset_id=asset_id,
                asset_name=asset_name,
                technologies=updated_technologies,
                repo_url=repo_url
            )
        except Exception as error:
            msg = "\U0001F4AC Something goes wrong when trying to send technologies to the CP, continuing... {error}".format(error=error)
            log_func(msg)

            response = None

        return response

def log_func(msg, new_line=True):
    click.echo(msg, nl=new_line, err=True)



from .version_searcher_result import VersionSearcherResult
from .timebased_version_seacher import TimeBasedVersionSearcher
from .sorted_by_versioning_style import SortedByVersioningStyle

__all__ = [
    'VersionSearcherResult',
    'TimeBasedVersionSearcher',
    'SortedByVersioningStyle',
]



class VersionSearcherResult(object):

    def __init__(self, **karg):
        self.current_tag = karg.get('current_tag')
        self.previous_tag = karg.get('previous_tag')
        self.current_commit = karg.get('current_commit')
        self.previous_commit = karg.get('previous_commit')

        if not (self.current_commit and self.previous_commit):
            raise ValueError(
                "The values of current_commit and previous_commit are required"
            )

    @property
    def current_version(self):
        return self.create_version_dict(
            self.current_commit,
            self.current_tag,
        )

    @property
    def previous_version(self):
        return self.create_version_dict(
            self.previous_commit,
            self.previous_tag,
        )

    @staticmethod
    def create_version_dict(commit, tag):
        return {
            'commit': commit,
            'tag': tag,
        }



import time
import traceback
from convisoappsec.logger import LOGGER, log_and_notify_ast_event


class RetryHandler:
    def __init__(self, flow_context=None, company_id=None, asset_id=None):
        self.max_retries = 5
        self.initial_delay = 1
        self.backoff_factor = 2
        self.flow_context = flow_context
        self.company_id = company_id
        self.asset_id = asset_id

    def execute_with_retry(self, func, *args, **kwargs):
        """Execute a function with retry logic and exponential backoff."""
        retries = 0
        delay = self.initial_delay

        while retries < self.max_retries:
            try:
                return func(*args, **kwargs)
            except Exception:
                retries += 1
                time.sleep(delay)
                delay *= self.backoff_factor

                if retries == self.max_retries:
                    full_trace = traceback.format_exc()
                    LOGGER.warning(
                        "⚠️ Maximum retries reached. Our technical team has been notified."
                    )

                    try:
                        log_and_notify_ast_event(
                            flow_context=self.flow_context, company_id=self.company_id, asset_id=self.asset_id,
                            ast_log=full_trace
                        )
                    except Exception as log_error:
                        LOGGER.warning(f"⚠️ Failed to log and notify AST event: {log_error}")



import jmespath
from convisoappsec.flow.graphql_api.beta.models.issues.iac import CreateIacFindingInput
from convisoappsec.flow.graphql_api.beta.models.issues.sast import CreateSastFindingInput
from convisoappsec.flow.graphql_api.beta.models.issues.sca import CreateScaFindingInput
from convisoappsec.flow.graphql_api.beta.models.issues.container import CreateOrUpdateContainerFindingInput
from convisoappsec.flow.graphql_api.beta.schemas import mutations
from convisoappsec.flow.graphql_api.v1.schemas import resolvers


class IssuesAPI(object):
    """ To operations on Issues's (aka, findings and vulnerabilities)) in Conviso Platform. """

    def __init__(self, conviso_graphql_client):
        self.__conviso_graphql_client = conviso_graphql_client

    def create_sast(self, sast_issue_model: CreateSastFindingInput):
        graphql_variables = {
            "input": sast_issue_model.to_graphql_dict()
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            mutations.CREATE_SAST_FINDING_INPUT,
            graphql_variables
        )

        expected_path = 'createSastFinding.issue'

        issue = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return issue

    def create_sca(self, sca_issue_model: CreateScaFindingInput):
        graphql_variables = {
            "input": sca_issue_model.to_graphql_dict()
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            mutations.CREATE_SCA_FINDING_INPUT,
            graphql_variables
        )

        expected_path = 'createScaFinding.issue'

        issue = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return issue

    def create_iac(self, issue_model: CreateIacFindingInput):
        graphql_variables = {
            "input": issue_model.to_graphql_dict()
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            mutations.CREATE_SAST_FINDING_INPUT,
            graphql_variables
        )

        expected_path = 'createSastFinding.issue'

        issue = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return issue

    def create_container(self, container_issue_model: CreateOrUpdateContainerFindingInput):
        graphql_variables = {
            "input": container_issue_model.to_graphql_dict()
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            mutations.CREATE_CONTAINER_FINDING_INPUT,
            graphql_variables
        )

        expected_path = 'createOrUpdateContainerFinding.issue'

        issue = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return issue

    def auto_close_vulnerabilities(self, company_id, asset_id, statuses, page=1, vulnerability_type=None):
        """ entry point for auto closing vulnerabilities on conviso platform """
        if vulnerability_type is None:
            vulnerability_type = ['SAST_FINDING', 'SCA_FINDING']

        graphql_variables = {
            'company_id': company_id,
            'asset_id': asset_id,
            'page': page,
            'per_page': 100,
            'statuses': statuses,
            'failure_types': vulnerability_type
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            resolvers.GET_ISSUES_FINGERPRINT,
            graphql_variables
        )

        expected_path = 'issues'

        issues = jmespath.search(
            expected_path,
            graphql_body_response
        )

        return issues

    def update_issue_status(self, issue_id, status, reason, control_sync_status_id):
        """ Update issue status on conviso platform """

        graphql_variables = {
            'issueId': issue_id,
            'status': status,
            'reason': reason,
            'controlSyncStatusId': control_sync_status_id
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            mutations.UPDATE_ISSUE_STATUS,
            graphql_variables
        )

        expected_path = 'changeIssueStatus.issue'

        issue = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return issue



import subprocess
import traceback
import click
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import DeployFormatter, PerformDeployException, asset_id_option
from convisoappsec.flowcli.deploy.create.context import pass_create_context
from convisoappsec.flowcli.deploy.create.with_.values import values
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from convisoappsec.flowcli.sast import sast
from convisoappsec.flowcli.sca import sca
from convisoappsec.flowcli.iac import iac
from convisoappsec.flowcli.vulnerability import vulnerability
from copy import deepcopy as clone
from convisoappsec.flow import GitAdapter
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.logger import LOGGER, log_and_notify_ast_event
from convisoappsec.common.cleaner import Cleaner


def get_default_params_values(cmd_params):
    """ Further information in https://click.palletsprojects.com/en/8.1.x/api/?highlight=params#click.Command.params

    Args:
        cmd_params (List[click.core.Parameter]):

    Returns:
        dict: default params values dictionarie
    """
    default_params = {}
    for param in cmd_params:
        unwanted = param.name in ['help', 'verbosity']
        if not unwanted:
            default_params.update({param.name: param.default})
    return default_params


def parse_params(ctx_params: dict, expected_params: list):
    """ Parse the params from the context extracting the expected params values to the context.

    Args:
        ctx_params (dict): context params: Further information at https://click.palletsprojects.com/en/8.1.x/api/?highlight=context%20param#click.Context.params
        expected_params (list): Further information at https://click.palletsprojects.com/en/8.1.x/api/?highlight=params#click.Command.params

    Returns:
        dict: parsed_params: parsed params as key and value
    """
    parsed_params = get_default_params_values(expected_params)
    for param in ctx_params:
        if param in parsed_params:
            parsed_params.update({param: ctx_params.get(param)})
    return parsed_params


def perform_sast(context, control_sync_status_id) -> int:
    """Setup and runs the "sast run" command.

    Args:
        context (<class 'click.core.Context'>): cloned context
        control_sync_status_id (str): control sync status id
    """
    sast_run = sast.commands.get('run')

    specific_params = {
        "deploy_id": context.obj.deploy['deploy_id'],
        "start_commit": context.obj.deploy['previous_commit'],
        "end_commit": context.obj.deploy['current_commit'],
        "control_sync_status_id": control_sync_status_id
    }
    context.params.update(specific_params)
    context.params = parse_params(context.params, sast_run.params)
    try:
        LOGGER.info(
            'Running SAST on deploy ID "{deploy_id}"...'
            .format(deploy_id=context.params["deploy_id"])
        )
        sast_run.invoke(context)

        return context.params.get('sast_vulnerability_count', 0)

    except Exception as err:
        raise click.ClickException(str(err)) from err


def perform_sca(context, control_sync_status_id) -> int:
    """Setup and runs the "sca run" command.

    Args:
        context (<class 'click.core.Context'>): cloned context
        control_sync_status_id (str): control sync status id
    """
    sca_run = sca.commands.get('run')
    context.params.update(
        {"deploy_id": context.obj.deploy['deploy_id'],
        "control_sync_status_id": control_sync_status_id}
    )

    context.params = parse_params(context.params, sca_run.params)
    try:
        LOGGER.info(
            'Running SCA on deploy ID "{deploy_id}"...'
            .format(deploy_id=context.params["deploy_id"])
        )
        sca_run.invoke(context)

        return context.params.get('sca_vulnerability_count', 0)

    except Exception as err:
        raise click.ClickException(str(err)) from err


def perform_iac(context, control_sync_status_id) -> None:
    """Setup and runs the "iac run" command.

    Args:
        context (<class 'click.core.Context'>): clonned context
        control_sync_status_id (str): control sync status id
    """
    iac_run = iac.commands.get('run')
    context.params.update(
        {"deploy_id": context.obj.deploy['deploy_id'],
        "control_sync_status_id": control_sync_status_id}
    )
    context.params = parse_params(context.params, iac_run.params)

    try:
        LOGGER.info(
            'Running IAC on deploy ID "{deploy_id}"...'
            .format(deploy_id=context.params["deploy_id"])
        )
        iac_run.invoke(context)
    except Exception as err:
        raise click.ClickException(str(err)) from err


def perform_vulnerabilities_service(context, company_id, control_sync_status_id) -> None:
    auto_close_run = vulnerability.commands.get('run')

    specific_params = {
        "deploy_id": context.obj.deploy['deploy_id'],
        "start_commit": context.obj.deploy['previous_commit'],
        "end_commit": context.obj.deploy['current_commit'],
        "company_id": context.params['company_id'] or company_id,
        "control_sync_status_id": control_sync_status_id
    }
    context.params.update(specific_params)
    context.params = parse_params(context.params, auto_close_run.params)

    try:
        LOGGER.info("[*] Verifying if any vulnerability was fixed...")
        auto_close_run.invoke(context)
    except Exception as err:
        raise click.ClickException(str(err)) from err


def perform_deploy(context, flow_context, prepared_context, control_sync_status_id):
    context.obj.output_formatter = DeployFormatter(format=DeployFormatter.DEFAULT)
    context.params = parse_params(context.params, values.params)
    repository_dir = context.params['repository_dir']
    asset_id = prepared_context.params.get('asset_id')

    if not asset_id:
        raise PerformDeployException("Asset ID is required")

    LOGGER.info("Creating new deploy...")
    try:

        created_deploy = values.invoke(context)

        if created_deploy is None:
            conviso_api = flow_context.create_conviso_graphql_client()
            conviso_api.control_sync_status.increase_count(
                control_sync_status_id=control_sync_status_id,
                asset_id=prepared_context.params['asset_id'],
                success_count=1 # increase by one just to be success
            )
            return None

        conviso_api = flow_context.create_conviso_graphql_client()
        api_key = flow_context.key
        git_adapter = GitAdapter(repository_dir)

        branch_name = get_branch_name(git_adapter, repository_dir)

        LOGGER.info(f"Creating deploy: asset_id={asset_id}, "
                    f"previous={created_deploy['previous_commit']}, "
                    f"current={created_deploy['current_commit']}")

        response = conviso_api.deploys.create_deploy(
            asset_id=asset_id,
            previous_commit=created_deploy['previous_commit'],
            current_commit=created_deploy['current_commit'],
            branch_name=branch_name,
            api_key=api_key,
            commit_history=created_deploy['commit_history']
        )

        response_deploy_id = response['createDeploy']['deploy']['id']
        deploy_params = {
            "deploy_id": response_deploy_id,
            "current_commit": created_deploy['current_commit'],
            "previous_commit": created_deploy['previous_commit'],
        }
        created_deploy.update(deploy_params)

        return created_deploy

    except Exception as err:
        error_message = str(err)
        error_text = "A deploy with the same previous and current commit already exists"        
        found = False

        if error_text in error_message:
            found = True
        elif hasattr(err, 'errors') and isinstance(err.errors, list):
            for nested_err in err.errors:
                if isinstance(nested_err, dict) and error_text in nested_err.get('message', ''):
                    found = True
                    break
                elif isinstance(nested_err, str) and error_text in nested_err:
                    found = True
                    break

        elif error_text in repr(err):
            found = True

        if found:
            LOGGER.warning("Deploy with same commits already exists")
            return None
        else:
            raise PerformDeployException(f"Failed to create deploy: {error_message}") from err


def get_branch_name(git_adapter, repository_dir):
    """Gets branch name"""
    try:
        return git_adapter.get_branch_name()
    except Exception as e:
        LOGGER.warning(f"HEAD is detached or error getting branch: {e}")
        LOGGER.info("Looking for most recent branch...")

        try:
            result = subprocess.run(
                ["git", "for-each-ref", "--sort=-creatordate",
                 "--format=%(refname:short)", "refs/heads/"],
                cwd=repository_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
                timeout=10
            )
            branches = result.stdout.decode().strip().splitlines()
            if branches:
                branch_name = branches[0]
                LOGGER.info(f"Using branch: {branch_name}")
                return branch_name
            else:
                LOGGER.warning("No branches found")
                return "unknown"
        except (subprocess.SubprocessError, subprocess.TimeoutExpired) as e:
            LOGGER.error(f"Error executing git command: {e}")
            return "unknown"


@click.command(
    context_settings=dict(
        allow_extra_args=True,
        ignore_unknown_options=True
    )
)
@asset_id_option(required=False)
@click.option(
    "--send-to-flow/--no-send-to-flow",
    default=True,
    show_default=True,
    required=False,
    help="""Enable or disable the ability of send analysis result
    reports to flow.""",
    hidden=True
)
@click.option(
    '-r',
    '--repository-dir',
    default=".",
    show_default=True,
    type=click.Path(exists=True, resolve_path=True),
    required=False,
    help="""The source code repository directory.""",
)
@click.option(
    "-c",
    "--current-commit",
    required=False,
    help="If no value is given the HEAD commit of branch is used. [DEPLOY]",
)
@click.option(
    "-p",
    "--previous-commit",
    required=False,
    help="""If no value is given, the value is retrieved from the lastest
    deploy at flow application. [DEPLOY]""",
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@click.option(
    '--vulnerability-auto-close',
    default=False,
    is_flag=True,
    help="Enable auto fixing vulnerabilities on cp.",
)
@click.option(
    '--cleanup',
    default=False,
    is_flag=True,
    show_default=True,
    help="Clean up system resources, including temporary files, stopped containers, unused Docker images and volumes.",
)
@help_option
@pass_flow_context
@pass_create_context
@click.pass_context
def run(context, create_context, flow_context, **kwargs):
    """ AST - Application Security Testing. Unifies deploy issue, SAST and SCA analyses.  """
    increase_failure_count = 1

    try:
        prepared_context = RequirementsVerifier.prepare_context(clone(context), from_ast=True)


        # After ensuring we have both asset and user permissions, AST is ready to start. We then create a control sync status
        # on the Conviso platform, which will appear on the scans page and in the recent scans list.
        conviso_api = flow_context.create_conviso_graphql_client()
        control_sync_status = conviso_api.control_sync_status.create_control_sync_status(
            asset_id=prepared_context.params['asset_id']
        )
        conviso_api.control_sync_status.update_control_sync_status(control_sync_status_id=control_sync_status['id'])

        try:
            prepared_context.obj.deploy = perform_deploy(
                clone(prepared_context), flow_context, prepared_context, control_sync_status['id']
            )

            if prepared_context.obj.deploy is None:
                return

            total_sast = perform_sast(clone(prepared_context), control_sync_status_id=control_sync_status['id'])
            total_sca = perform_sca(clone(prepared_context), control_sync_status_id=control_sync_status['id'])
            perform_iac(clone(prepared_context), control_sync_status_id=control_sync_status['id'])

            total_vulnerability_count = total_sast + total_sca

            company_id = prepared_context.params['company_id']

            if context.params['vulnerability_auto_close'] is True:
                try:
                    perform_vulnerabilities_service(clone(prepared_context), company_id, control_sync_status['id'])
                except Exception:
                    LOGGER.info("An issue occurred while attempting to fix vulnerabilities. Our technical team has been notified.")
                    full_trace = traceback.format_exc()
                    log_and_notify_ast_event(flow_context=flow_context, company_id=company_id,
                                             asset_id=prepared_context.params['asset_id'], ast_log=full_trace)
                    return

            if context.params.get('cleanup'):
                try:
                    LOGGER.info("🧹 Cleaning up ...")
                    cleaner = Cleaner()
                    cleaner.cleanup()
                except Exception as e:
                    LOGGER.info(f"An error occurred while cleaning up. Our technical team has been notified.")
                    full_trace = traceback.format_exc()
                    log_and_notify_ast_event(
                        flow_context=flow_context, company_id=company_id, asset_id=prepared_context.params['asset_id'],
                        ast_log=full_trace
                    )
                    return

            conviso_api.control_sync_status.update_control_sync_status(
                control_sync_status_id=control_sync_status['id'],
                external_vulnerability_count=total_vulnerability_count
            )

            # increase success_count by one just to match external_vulnerability_count when ast runs successfully and if
            # success count and external_vulnerability_count are equals, the scan status will be moved from pending to success.
            conviso_api.control_sync_status.increase_count(
                control_sync_status_id=control_sync_status['id'],
                asset_id=prepared_context.params['asset_id'],
                success_count=total_vulnerability_count
            )

        except Exception as err:
            failure_details = f"{str(err)}, {traceback.format_exc()}"

            conviso_api.control_sync_status.increase_count(
                control_sync_status_id=control_sync_status['id'],
                asset_id=prepared_context.params['asset_id'],
                failure_count=increase_failure_count,
                failure_reason=str(failure_details)
            )
            raise click.ClickException(str(err)) from err

    except Exception as err:
        error_message = str(err)

        if "A deploy with the same previous and current commit already exists" in error_message:
            LOGGER.warning("Deploy with same commits already exists")
            return None
        else:
            LOGGER.error(f"AST initialization failed. Please contact support with the following error: {err}")
            raise click.ClickException(str(err)) from err


@click.group()
def ast():
    pass


ast.add_command(run)



import requests

from convisoappsec.common.graphql.error_handlers import GraphQlErrorHandler, RequestErrorHandler
from convisoappsec.version import __version__


class GraphQLClient:
    DEFAULT_HEADERS = {
        'Accept': 'application/json',
        'Content-Type': 'application/json',
        "User-Agent": "AST:{version}".format(version=__version__),
    }

    def __init__(self, url, headers={}):
        self.url = url
        self.__session = requests.Session()
        self.__session.headers.update(
            **self.DEFAULT_HEADERS,
            **headers
        )

    def execute(self, query, variables={}):
        try:
            payload = self._build_graphql_payload(query, variables)

            response = self.__session.post(
                url=self.url,
                json=payload,
            )

            response.raise_for_status()

        except Exception as e:
            handler = RequestErrorHandler(e)
            handler.handle_request_error()

        json_response = response.json()
        graphql_handler = GraphQlErrorHandler(json_response)
        graphql_handler.raise_on_graphql_error()
        graphql_handler.raise_on_graphql_body_error()

        return json_response['data']

    @staticmethod
    def _build_graphql_payload(query, variables):
        data = {
            'query': query,
            'variables': variables
        }

        return data



class Error(Exception):
    pass


class AuthenticationError(Error):
    pass


class ServerError(Error):
    pass


class ResponseError(Error):
    def __init__(self, message, code=None):
        super().__init__(message)
        self.code = code



import docker
from contextlib import suppress
import tempfile
from uuid import uuid4

from convisoappsec.flow.util import SourceCodeCompressor
from .exceptions import SourceCodeScannerException


class SourceCodeScanner(object):
    SUCCESS_EXIT_CODE = 0
    '''
    hooks:
      def _pre_pull(self):
          :return: void

      def _capture_stdout(self, stdout_bytes)
        :param stdout_bytes: chunks generated by stdout
        :paramtype stdout_bytes: bytes
        :return: void
    _pre_scan
    _scan_stdout
    _post_scan
    '''

    def __init__(self, source_code_dir, create_source_code_volume = True):
        uuid = str(uuid4())
        self.docker = docker.from_env(version="auto")
        self.__container_name = "source_code_scanner_{0}".format(
            uuid
        )
        self.__source_code_dir = source_code_dir
        self.__create_source_code_volume = create_source_code_volume

        if self.__create_source_code_volume:
            self.__source_code_volume_name = "source_code_scanner_src_{0}".format(
                uuid
            )
        else:
            self.__source_code_volume_name = None

    @property
    def repository(self):
        raise Exception('Not implemented yet!')

    @property
    def tag(self):
        raise Exception('Not implemented yet!')

    @property
    def command(self):
        raise Exception('Not implemented yet!')

    @property
    def container_source_dir(self):
        raise Exception('Not implemented yet!')

    @property
    def image(self):
        return "{repository}:{tag}".format(
            repository=self.repository,
            tag=self.tag,
        )

    @property
    def volumes(self):
        if not self.__create_source_code_volume:
            return {}

        return {
            self.__source_code_volume_name: {
                'bind': self.container_source_dir,
                'mode': 'rw',
            }
        }

    def __get_container(self):
        return self.docker.containers.get(
            self.__container_name
        )

    def __get_source_code_volume(self):
        return self.docker.volumes.get(
            self.__source_code_volume_name
        )

    @property
    def __container(self):
        try:
            return self.__get_container()
        except docker.errors.NotFound:
            return self.__create_container()

    def __create_container(self):
        return self.docker.containers.create(
            self.image,
            name=self.__container_name,
            volumes=self.volumes,
            detach=True,
            command=self.command,
            working_dir=self.container_source_dir if self.container_source_dir != '/code' else '/code'
        )

    def __pull_image(self):
        if self.has_pre_pull:
            self._pre_pull()

        self.docker.images.pull(self.repository, self.tag)

    def __load_source_code(self):
        container = self.__container

        with tempfile.TemporaryFile() as fileobj:
            compressor = SourceCodeCompressor(
                self.__source_code_dir
            )

            compressor.write_to(fileobj)
            fileobj.seek(0)

            container.put_archive(
                self.container_source_dir,
                fileobj
            )

    def scan(self):
        self.__pull_image()
        self.__load_source_code()
        container = self.__container

        container.start()

        if self.has_read_scan_stderr:
            self._read_scan_stderr(
                container.logs(
                    stream=True, stdout=False, stderr=True
                )
            )

        if self.has_read_scan_stdout:
            self._read_scan_stdout(
                container.logs(
                    stream=True, stdout=True, stderr=False
                )
            )

        wait_result = container.wait()
        status_code = wait_result.get('StatusCode')

        if not status_code == self.SUCCESS_EXIT_CODE:
            raise SourceCodeScannerException(
                'Source code scanning fail'
            )

    def __has_method(self, method_name):
        return hasattr(self, method_name)

    @property
    def has_read_scan_stdout(self):
        return self.__has_method('_read_scan_stdout')

    @property
    def has_read_scan_stderr(self):
        return self.__has_method('_read_scan_stderr')

    @property
    def has_pre_pull(self):
        return self.__has_method('_pre_pull')

    def __del__(self):
        with suppress(Exception):
            container = self.__get_container()
            container.remove()

        with suppress(Exception):
            source_code_volume = self.__get_source_code_volume()
            source_code_volume.remove()



import subprocess
import shutil
import os
from pathlib import Path
from typing import List, Optional


class RunnerCleanup:
    """Class to handle CI/CD runner cleanup operations"""

    def __init__(self, temp_dirs: Optional[List[str]] = None):
        """
        Initialize cleanup class

        Args:
            temp_dirs: Optional list of temporary directories to clean
        """
        self.temp_dirs = temp_dirs or [
            "/tmp",
            "/var/tmp",
            str(Path.home() / ".cache")
        ]
        self.is_runner = self._is_running_in_ci()
        self._check_requirements()

    def _is_running_in_ci(self) -> bool:
        """Check if running in a CI/CD environment"""
        ci_indicators = {
            'CI': None,
            'GITLAB_CI': None,
            'GITHUB_ACTIONS': None,
            'JENKINS_URL': None,
            'TRAVIS': None,
            'CIRCLECI': None,
            'BUILD_ID': None,
            'CI_JOB_ID': None,
            'GITHUB_RUN_ID': None,
        }

        # Check if any CI-specific environment variable is set
        for var in ci_indicators:
            if os.environ.get(var):
                print(f"Detected CI environment: {var}")
                return True

        # Additional check for runner-specific variables
        if os.environ.get('RUNNER_TEMP') or os.environ.get('RUNNER_WORKSPACE'):
            print("Detected GitHub Actions runner environment")
            return True

        print("No CI environment detected")
        return False

    def _run_command(self, command: str) -> Optional[str]:
        """Execute shell command and return output"""

        try:
            result = subprocess.run(
                command,
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True
            )
            return result.stdout.strip()
        except subprocess.CalledProcessError as e:
            print(f"Error executing '{command}': {e.stderr}")
            return None

    def _check_requirements(self) -> None:
        """Check system requirements"""
        if not self._run_command("docker --version"):
            raise RuntimeError("Docker is not installed or not accessible")

    def get_disk_space(self) -> tuple[float, float, float]:
        """Get disk space information in GB"""
        stat = shutil.disk_usage("/")
        total = stat.total / (1024 ** 3)
        used = stat.used / (1024 ** 3)
        free = stat.free / (1024 ** 3)
        return total, used, free

    def cleanup_docker(self) -> None:
        """Clean up Docker resources"""
        commands = [
            "docker container prune -f",
            "docker image prune -f",
            "docker volume prune -f",
            "docker network prune -f"
        ]

        for cmd in commands:
            self._run_command(cmd)

    def cleanup_temp(self) -> None:
        """Clean up temporary directories"""

        for dir_path in self.temp_dirs:
            dir_path = Path(dir_path)
            if not dir_path.exists():
                continue

            for item in dir_path.iterdir():
                try:
                    if item.is_file():
                        item.unlink()
                    elif item.is_dir():
                        shutil.rmtree(item, ignore_errors=True)
                except Exception as e:
                    print(f"Failed to remove {item}: {e}")

    def cleanup_all(self):
        """Perform full cleanup and return before/after disk space stats"""
        print("Cleaning up ...")

        self.cleanup_docker()
        self.cleanup_temp()



CREATE_ASSET = """ 
mutation (
  $companyId: Int!,
  $name: String!,
  $scanType: AssetScan!
) {
  createAsset(
    input: {
      companyId: $companyId,
      name: $name,
      scanType: $scanType
    }
  ) {
    asset {
      id
      name
      createdAt
    }
    errors
  }
}
"""

UPDATE_ASSET = """
mutation (
  $id: ID!,
  $companyId: Int!,
  $name: String!,
  $tecnologyList: [String!],
  $repoUrl: String
) {
  updateAsset(
    input: {
      id: $id,
      companyId: $companyId,
      name: $name,
      tecnologyList: $tecnologyList,
      repoUrl: $repoUrl
    }
  ) {
    asset {
      id
    }
  }
}
"""

IMPORT_SBOM = """
mutation (
  $file: Upload!,
  $assetId: ID!,
  $companyId: ID!
) {
  importSbom(
    input: {
      file: $file,
      assetId: $assetId,
      companyId: $companyId
    }
  ) {
    success
  }
}
"""

IMPORT_CONTAINER = """
mutation (
  $file: Upload!,
  $assetId: ID!,
  $companyId: ID!
) {
  importContainerFindingsFile(
    input: {
      file: $file,
      assetId: $assetId,
      companyId: $companyId
    }
  ) {
    success
  }
}
"""

LOG_AST_ERROR = """
mutation (
  $companyId: ID!,
  $assetId: ID!,
  $log: String!
) {
  logAstError(
    input: {
      companyId: $companyId,
      assetId: $assetId,
      log: $log
    }
  ) {
    success
  }
}
"""

CREATE_DEPLOY = """
mutation (
  $assetId: ID!,
  $previousCommit: String!,
  $currentCommit: String!,
  $branchName: String,
  $diffContent: Upload!,
  $commitHistory: Upload!
) {
  createDeploy(
    input: {
      assetId: $assetId,
      previousCommit: $previousCommit,
      currentCommit: $currentCommit,
      branchName: $branchName,
      diffContent: $diffContent,
      commitHistory: $commitHistory
    }
  ) {
    deploy {
      id
    }
  }
}
"""

IMPORT_FINDINGS = """
mutation (
  $file: Upload!,
  $assetId: ID!,
  $companyId: ID!
  $vulnerabilityTypes: [Issue!]!
  $deployId: ID
  $commitRef: String
  $controlSyncStatusId: ID
) {
  importAstFindingsFile(
    input: {
      file: $file,
      assetId: $assetId,
      companyId: $companyId,
      vulnerabilityTypes: $vulnerabilityTypes,
      deployId: $deployId,
      commitRef: $commitRef
      controlSyncStatusId: $controlSyncStatusId
    }
  ) {
    success
  }
}
"""

CREATE_CONTROL_SYNC_STATUS = """
mutation ($assetId: ID!) {
  createControlSyncStatus(
    input: {
      assetId: $assetId,
      integration: CONVISO_AST,
      externalVulnerabilityCount: 0
    }
  ) {
    controlSyncStatus {
      id
    }
    success
  }
}
"""

UPDATE_CONTROL_SYNC_STATUS = """
mutation (
    $id: ID!
    $externalVulnerabilityCount: Int!
) {
  updateControlSyncStatus(
    input: {
      id: $id
      externalVulnerabilityCount: $externalVulnerabilityCount
    }
  ) {
    controlSyncStatus {
      id
    }
  }
}
"""

INCREASE_ISSUE_SCAN_COUNT = """
mutation (
  $controlSyncStatusId: ID!
  $assetId: ID!
  $successCount: Int
  $failureCount: Int
  $failureReason: String
) {
  increaseIssueScanCount(
    input: {
      controlSyncStatusId: $controlSyncStatusId
      assetId: $assetId
      successCount: $successCount
      failureCount: $failureCount
      failureReason: $failureReason
      integration: CONVISO_AST
    }
  ) {
    controlSyncStatus {
	  id
	}
  }
}
"""



import click
import click_log
import logging
import yaml
import json
import jsonschema
from datetime import datetime, timedelta
from pkg_resources import resource_filename
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.flowcli.common import asset_id_option
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from convisoappsec.flowcli.companies.ls import Companies

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


@click.command('assert-security-rules')
@click_log.simple_verbosity_option(logger)
@asset_id_option(
    required=False
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    "-r",
    "--repository-dir",
    default=".",
    show_default=True,
    type=click.Path(
        exists=True,
        resolve_path=True,
    ),
    required=False,
    help="The source code repository directory.",
)
@click.option(
    '--rules-file',
    'rules_file',
    type=click.File('r'),
    required=True
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@help_option
@pass_flow_context
@click.pass_context
def assert_security_rules(
        context, flow_context, asset_id, company_id, repository_dir, rules_file, asset_name
):
    prepared_context = RequirementsVerifier.prepare_context(context)
    asset_id = prepared_context.params['asset_id']

    if company_id is None:
        companies = Companies()
        company = companies.ls(flow_context, company_id=company_id)
        company_id = company[0]['id']

    try:
        rules = yaml.load(
            rules_file,
            Loader=yaml.Loader
        )

        click.secho(
            '💬 Starting vulnerability security rules assertion...',
            bold=True
        )

        click.secho(
            "💬 Applying the given rules at the security gate:\n{0}".format(yaml.dump(rules)),
            bold=True
        )

        conviso_api = flow_context.create_conviso_graphql_client()
        statuses = ['IDENTIFIED', 'IN_PROGRESS', 'AWAITING_VALIDATION']

        if validate_json(rules)[0] is False:
            msg = click.secho(
                '💬 Error: Validation of the security gate YAML file failed during the validation step!',
                bold=True
            )

            return msg

        current_date = datetime.now()
        default_end_date = datetime(
            current_date.year, current_date.month, current_date.day, 23, 59, 59
        )

        tolerated_days_exist = any(
            'max_days_to_fix' in severity for rule in rules['rules'] for severity in rule['severity'].values()
        )

        issues_set = set()

        if tolerated_days_exist:
            tolerated_days_and_severity = tolerated_days_by_severity(rules)
            for severity, days in tolerated_days_and_severity.items():
                end_date = default_end_date - timedelta(days=days)
                end_date = end_date.isoformat() + "Z"
                severity_issues = conviso_api.issues.get_issues_stats(
                    asset_id, company_id, statuses, end_date=end_date
                )
                for issue in severity_issues:
                    severity = severity.upper()

                    if issue['value'] == severity:
                        issues_set.add((severity, issue["count"]))

            issues = [{'value': value, 'count': count} for value, count in issues_set]
        else:
            issues = conviso_api.issues.get_issues_stats(asset_id, company_id, statuses)

        response = validate_rules(issues, rules)

        __raise_if_gate_locked(response)

        click.secho(
            '✅ Vulnerability security rules assertion finished.',
            bold=True
        )
    except Exception as e:
        raise click.ClickException(str(e)) from e


def tolerated_days_by_severity(rules):
    """
    Returns tolerated days for each severity to facilitate validation.
    """
    days_by_severity = {}

    for rule in rules['rules']:
        for severity, values in rule['severity'].items():
            if 'max_days_to_fix' in values:
                days_by_severity[severity] = values['max_days_to_fix']
            else:
                days_by_severity[severity] = 0

    return days_by_severity


def __raise_if_gate_locked(response):
    if response['locked']:
        click.secho('💬 Vulnerabilities summary...', bold=True)

        logger.info(
            json.dumps(response['summary'], indent=4)
        )

        raise click.ClickException(
            'Vulnerabilities quantity offending security rules.'
        )


def validate_rules(issues, rules):
    """ function to validate security gate rules """
    response = {"locked": False, "summary": [{"from": "any", "severity": {}}]}
    parsed_issues = {issue['value']: issue['count'] for issue in issues}

    for i, rule in enumerate(rules['rules']):
        for criticity, rule_max in rule['severity'].items():
            if parsed_issues[criticity.upper()] > rule_max['maximum']:
                response['locked'] = True
            response["summary"][i]["severity"].update({criticity: {"quantity": parsed_issues[criticity.upper()]}})

    return response


def validate_json(rules):
    """ Validate a JSON document against a JSON schema. """
    schema_path = resource_filename('convisoappsec', 'flowcli/vulnerability/rules_schema.json')

    with open(schema_path, 'r') as json_file:
        schema = json.load(json_file)
    try:
        jsonschema.validate(rules, schema)
        return True, "Validation successful"
    except jsonschema.exceptions.ValidationError as e:
        return False, str(e)


EPILOG = '''
'''

SHORT_HELP = ''

command = 'conviso vulnerability assert-security-rules'
assert_security_rules.short_help = SHORT_HELP
assert_security_rules.epilog = EPILOG.format(
    command=command,
)



import click
import json
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.sast.sastbox import SASTBox
from convisoappsec.flowcli.common import (asset_id_option, on_http_error)
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from convisoappsec.flow import GitAdapter
from convisoappsec.common.box import ContainerWrapper
from convisoappsec.common.graphql.errors import ResponseError
from convisoappsec.flow.cleaner import RunnerCleanup
from convisoappsec.logger import LOGGER


@click.command()
@asset_id_option(required=False)
@click.option(
    "-s",
    "--start-commit",
    required=False,
    help="If no value is set so the empty tree hash commit is used.",
)
@click.option(
    "-e",
    "--end-commit",
    required=False,
    help="""If no value is set so the HEAD commit
    from the current branch is used""",
)
@click.option(
    "-r",
    "--repository-dir",
    default=".",
    show_default=True,
    type=click.Path(
        exists=True,
        resolve_path=True,
    ),
    required=False,
    help="The source code repository directory.",
)
@click.option(
    "--fail-on-severity-threshold",
    required=False,
    help="If the threshold of the informed severity and higher has reach, then the command will fail after send the results to AppSec Flow.\n \
    The severity levels are: UNDEFINED, INFO, LOW, MEDIUM, HIGH, CRITICAL.",
    type=click.Tuple([str, int]),
    default=(None, None),
)
@click.option(
    "--fail-on-threshold",
    required=False,
    help="If the threshold has reach then the command will fail after send the result to AppSec Flow",
    type=int,
    default=False,
)
@click.option(
    "--send-to-flow/--no-send-to-flow",
    default=True,
    show_default=True,
    required=False,
    hidden=True,
    help="""Enable or disable the ability of send analysis result
    reports to flow.""",
)
@click.option(
    "--deploy-id",
    default=None,
    required=False,
    hidden=True,
    envvar=("CONVISO_DEPLOY_ID", "FLOW_DEPLOY_ID")
)
@click.option(
    "--sastbox-registry",
    default="",
    required=False,
    hidden=True,
    envvar=("CONVISO_SASTBOX_REGISTRY", "FLOW_SASTBOX_REGISTRY"),
)
@click.option(
    "--sastbox-repository-name",
    default="",
    required=False,
    hidden=True,
    envvar=("CONVISO_SASTBOX_REPOSITORY_NAME", "FLOW_SASTBOX_REPOSITORY_NAME"),
)
@click.option(
    "--sastbox-tag",
    default=SASTBox.DEFAULT_TAG,
    required=False,
    hidden=True,
    envvar=("CONVISO_SASTBOX_TAG", "FLOW_SASTBOX_TAG"),
)
@click.option(
    "--sastbox-skip-login/--sastbox-no-skip-login",
    default=False,
    required=False,
    hidden=True,
    envvar=("CONVISO_SASTBOX_SKIP_LOGIN", "FLOW_SASTBOX_SKIP_LOGIN"),
)
@click.option(
    '--experimental',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable experimental features.",
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@click.option(
    '--vulnerability-auto-close',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable auto fixing vulnerabilities on cp.",
)
@click.option(
    '--from-ast',
    default=False,
    is_flag=True,
    hidden=True,
    help="Internal use only.",
)
@click.option(
    '--cleanup',
    default=False,
    is_flag=True,
    show_default=True,
    help="Clean up system resources, including temporary files, stopped containers, unused Docker images and volumes.",
)
@click.option(
    '--control-sync-status-id',
    required=False,
    hidden=True,
    help="Control sync status id.",
)
@help_option
@pass_flow_context
@click.pass_context
def run(context, flow_context, asset_id, company_id, end_commit, start_commit, repository_dir, send_to_flow, deploy_id,
        sastbox_registry, sastbox_repository_name, sastbox_tag, sastbox_skip_login, fail_on_threshold,
        fail_on_severity_threshold, experimental, asset_name, vulnerability_auto_close, from_ast, cleanup,
        control_sync_status_id):

    if context.params.get('cleanup'):
        try:
            log_func("Cleaning before runs auto close ...")
            cleaner = RunnerCleanup()
            cleaner.cleanup_all()
        except Exception as e:
            log_func(f"Failed to clean, trying any way ... Error: {str(e)}")

    if not from_ast:
        prepared_context = RequirementsVerifier.prepare_context(context)

        params_to_copy = [
            'asset_id', 'company_id', 'start_commit', 'end_commit',
            'repository_dir', 'send_to_flow', 'deploy_id', 'sastbox_registry',
            'sastbox_repository_name', 'sastbox_tag', 'sastbox_skip_login',
            'experimental', 'asset_name', 'vulnerability_auto_close', 'company_id'
        ]

        for param_name in params_to_copy:
            context.params[param_name] = (
                    locals()[param_name] or prepared_context.params[param_name]
            )

    conviso_api = flow_context.create_conviso_api_client_beta()
    company_id = company_id or context.params['company_id']
    asset_id = context.params['asset_id']
    statuses = ['CREATED', 'IDENTIFIED', 'IN_PROGRESS', 'AWAITING_VALIDATION', 'FIX_ACCEPTED']

    page = 1
    merged_issues_sast = []
    merged_issues_sca = []

    log_func("Running automatic closure of vulnerabilities...")

    while True:
        try:
            issues_from_cp = conviso_api.issues.auto_close_vulnerabilities(
                company_id, asset_id, statuses, page
            )
        except ResponseError as error:
            if 'Variable $company_id' in str(error):
                log_func(f"Invalid company_id passed: {company_id}", fg='red')
            else:
                log_func(f"error: {error}", fg='red')

            log_func("⚠️ Auto-close will not be performed at this time. Please set it using --company-id and try again.", fg='red')
            return

        total_pages = issues_from_cp['metadata']['totalPages']
        issues_collection = issues_from_cp['collection']
        issues_collection = [item for item in issues_collection if item['scanSource'] == 'conviso_ast']
        sast_issues = [item for item in issues_collection if item['type'] == 'SAST_FINDING']
        sca_issues = [item for item in issues_collection if item['type'] == 'SCA_FINDING']

        # could not be if and else because on the response could contain both.
        if sast_issues:
            merged_issues_sast.extend(sast_issues)

        if sca_issues:
            merged_issues_sca.extend(sca_issues)

        if total_pages == 0 or total_pages == page:
            break
        else:
            page += 1

    # sast issues filter
    sast_issues_with_fix_accepted = [item for item in merged_issues_sast if item['status'] == 'FIX_ACCEPTED']
    sast_issues_without_fix_accepted = [item for item in merged_issues_sast if item['status'] != 'FIX_ACCEPTED']

    # sca issues filter
    sca_issues_with_fix_accepted = [item for item in merged_issues_sca if item['status'] == 'FIX_ACCEPTED']
    sca_issues_without_fix_accepted = [item for item in merged_issues_sca if item['status'] != 'FIX_ACCEPTED']

    if len(issues_from_cp) == 0:
        log_func("No vulnerabilities were found on the Conviso Platform!")
        return

    # Starting executing the ast again
    sast_hash_issues = perform_sastbox_scan(sastbox_registry, sastbox_repository_name, sastbox_tag, repository_dir)

    sca_hash_issues = perform_sca_scan(repository_dir=repository_dir)
    iac_hash_issues = perform_iac_scan(repository_dir=repository_dir)

    # we need to append the two lists because at the moment this was made, iac and sast has sast as type on cp.
    sast_hash_issues = sast_hash_issues + iac_hash_issues
    # end ast execution

    set_of_sast_hash_issues = set(sast_hash_issues)
    set_of_sca_hash_issues = set(sca_hash_issues)

    close_sast_issues(conviso_api, sast_issues_without_fix_accepted, set_of_sast_hash_issues, control_sync_status_id)
    close_sca_issues(conviso_api, sca_issues_without_fix_accepted, set_of_sca_hash_issues, control_sync_status_id)

    sast_issues_to_reopen = [
        {'id': item['id'], 'originalIssueIdFromTool': item['originalIssueIdFromTool']}
        for item in sast_issues_with_fix_accepted if item['originalIssueIdFromTool'] in sast_hash_issues
    ]

    sca_issues_to_reopen = [
        {'id': item['id'], 'originalIssueIdFromTool': item['originalIssueIdFromTool']}
        for item in sca_issues_with_fix_accepted if item['originalIssueIdFromTool'] in sca_hash_issues
    ]

    if sast_issues_to_reopen:
        log_func("SAST: reopening {issues} vulnerability/vulnerabilities on conviso platform ...".format(
            issues=len(sast_issues_to_reopen))
        )

        reopen_issues(conviso_api, sast_issues_to_reopen, control_sync_status_id)

    if sca_issues_to_reopen:
        log_func("SCA: reopening {issues} vulnerability/vulnerabilities on conviso platform ...".format(
            issues=len(sca_issues_to_reopen))
        )

        reopen_issues(conviso_api, sca_issues_to_reopen, control_sync_status_id)


def close_sast_issues(conviso_api, issues_from_cp, issues_from_current_scan, control_sync_status_id):
    """
    method to close sast issues on conviso platform

    # issues_from_cp are issues already on conviso platform.
    # issues_from_current_scan are issues identify on each time the ast command runs, these are always from a full code
    base scan
    """
    log_func("SAST: Verifying if any vulnerability was solved...")

    differences = [
        {'id': item['id'], 'originalIssueIdFromTool': item['originalIssueIdFromTool']}
        for item in issues_from_cp if item['originalIssueIdFromTool'] not in issues_from_current_scan
    ]

    if len(differences) == 0:
        log_func("No vulnerabilities have been fixed yet...")
        return

    log_func("SAST: Fixing {issues} vulnerabilities on the Conviso Platform...".format(issues=len(differences)))

    for issue in differences:
        issue_id = issue['id']
        status = 'FIX_ACCEPTED'
        reason = ("The vulnerability is no longer found in the specified file/path. Its status has been updated by "
                  "Conviso AST")

        conviso_api.issues.update_issue_status(
            issue_id=issue_id, status=status, reason=reason, control_sync_status_id=control_sync_status_id
        )


def close_sca_issues(conviso_api, issues_from_cp, issues_from_current_scan, control_sync_status_id):
    """ method to close sca issues on conviso platform """

    log_func("SCA: Verifying if any vulnerability was solved...")
    differences = [
        {'id': item['id'], 'originalIssueIdFromTool': item['originalIssueIdFromTool']}
        for item in issues_from_cp if item['originalIssueIdFromTool'] not in issues_from_current_scan
    ]

    if len(differences) == 0:
        log_func("No vulnerabilities have been fixed yet...")
        return

    log_func("SCA: Fixing {issues} vulnerabilities on the Conviso Platform...".format(issues=len(differences)))

    for issue in differences:
        issue_id = issue['id']
        status = 'FIX_ACCEPTED'
        reason = ("The vulnerability is no longer found in the specified file/path. Its status has been updated by "
                  "Conviso AST")

        conviso_api.issues.update_issue_status(
            issue_id=issue_id, status=status, reason=reason, control_sync_status_id=control_sync_status_id
        )


def reopen_issues(conviso_api, issues, control_sync_status_id):
    """ Method to reopen issues when was detected again after a full scan is performed """

    for issue in issues:
        issue_id = issue['id']
        status = 'IDENTIFIED'
        reason = 'Status has been updated from Fixed to Identified by Conviso AST'

        conviso_api.issues.update_issue_status(
            issue_id=issue_id, status=status, reason=reason, control_sync_status_id=control_sync_status_id
        )

def perform_sastbox_scan(sastbox_registry, sastbox_repository_name, sastbox_tag, repository_dir):
    LOGGER.info('   🔍 [SAST] Running scan...')

    sastbox = SASTBox(registry=sastbox_registry, repository_name=sastbox_repository_name, tag=sastbox_tag)
    git_adapter = GitAdapter(repository_dir, unshallow_repository=False)

    reports = sastbox.run_scan_diff(
        repository_dir,
        git_adapter.head_commit,
        git_adapter.empty_repository_tree_commit
    )

    try:
        first_report = next(reports)
        with open(str(first_report), 'r') as f:
            issues = json.load(f).get("issues", [])

        return [
            issue.get("hash_issue") or issue.get("hash_issue_v2")
            for issue in issues
            if issue.get("hash_issue") or issue.get("hash_issue_v2")
        ]

    except (StopIteration, Exception) as e:
        if isinstance(e, StopIteration):
            return []
        return []

@pass_flow_context
def perform_sca_scan(flow_context, repository_dir):
    LOGGER.info('   🔍 [SCA] Running scan...')

    try:
        REQUIRED_CODEBASE_PATH = '/code'
        OSV_SCANNER_IMAGE_NAME = 'osv_scanner'

        scanners = {
            OSV_SCANNER_IMAGE_NAME: {
                'repository_name': OSV_SCANNER_IMAGE_NAME,
                'tag': 'latest',
                'command': [
                    '-c', REQUIRED_CODEBASE_PATH,
                    '-f', 'json',
                    '-o', '/{}.json'.format(OSV_SCANNER_IMAGE_NAME)
                ],
                'repository_dir': repository_dir
            },
        }

        conviso_rest_api = flow_context.create_conviso_rest_api_client()
        token = conviso_rest_api.docker_registry.get_sast_token()
        scabox = ContainerWrapper(
            token=token,
            containers_map=scanners,
            logger=None,
            timeout=7200
        )

        scabox.run()

        results_filepaths = []
        for unit in scabox.scanners:
            file_path = unit.results
            if file_path:
                results_filepaths.append(file_path)

        hash_issues = []

        for report_path in results_filepaths:
            try:
                with open(report_path, 'r') as report_file:
                    report_content = json.load(report_file)
                    issues = report_content.get("issues", [])
                    hash_issues.extend(issue.get("hash_issue") for issue in issues)

            except (FileNotFoundError, json.JSONDecodeError) as e:
                print(f"Error processing {report_path}: {e}")

        return hash_issues

    except Exception as e:
        on_http_error(e)
        raise click.ClickException(str(e)) from e


@pass_flow_context
def perform_iac_scan(flow_context, repository_dir):
    """ Perform an iac scan """
    LOGGER.info('   🔍 [IaC] Running scan...')

    try:
        REQUIRED_CODEBASE_PATH = '/code'
        IAC_IMAGE_NAME = 'iac_scanner_checkov'
        IAC_SCAN_FILENAME = '/{}.json'.format(IAC_IMAGE_NAME)
        containers_map = {
            IAC_IMAGE_NAME: {
                'repository_dir': repository_dir,
                'repository_name': IAC_IMAGE_NAME,
                'tag': 'unstable',
                'command': [
                    '-c', REQUIRED_CODEBASE_PATH,
                    '-o', IAC_SCAN_FILENAME,
                ],
            },
        }

        conviso_rest_api = flow_context.create_conviso_rest_api_client()
        token = conviso_rest_api.docker_registry.get_sast_token()
        scanners_wrapper = ContainerWrapper(
            token=token,
            containers_map=containers_map,
            logger=None,
            timeout=7200
        )

        scanners_wrapper.run()

        results_filepaths = []  # [str(r.results) for r in scanners_wrapper.scanners]
        hash_issues = []
        for r in scanners_wrapper.scanners:
            report_filepath = r.results
            if report_filepath:
                results_filepaths.append(report_filepath)

        for report_path in results_filepaths:
            try:
                with open(report_path) as report_file:
                    report_content = json.load(report_file)
                    results = report_content.get("runs", [])[0].get("results", [])
                    for result in results:
                        partial_fingerprints = result.get("partialFingerprints", {})
                        hash_issue = partial_fingerprints.get("hashIssueV2")
                        if hash_issue:
                            hash_issues.append(hash_issue)
            except (FileNotFoundError, json.JSONDecodeError) as e:
                print(f"Error processing {report_path}: {e}")

        return hash_issues
    except Exception as e:
        on_http_error(e)
        raise click.ClickException(str(e)) from e

def log_func(msg, new_line=True, fg='blue'):
    click.echo(click.style(msg, bold=True, fg=fg), nl=new_line, err=True)



import os
import platform
import subprocess

class EnvironmentChecker:
    """
    A class to check the system environment for required software and services.
    """

    def __init__(self):
        self.checks = {}

    def _run_command(self, command, check_running=False):
        """Helper function to run shell commands and capture output/errors."""
        try:
            process = subprocess.run(command, capture_output=True, text=True, check=not check_running)
            if check_running:
                return process.returncode == 0, "", ""
            return process.returncode == 0, process.stdout.strip(), process.stderr.strip()
        except FileNotFoundError:
            return False, "", "Command not found"
        except subprocess.CalledProcessError as e:
            return False, "", e.stderr.strip()

    def check_docker(self):
        """Checks if Docker is installed and the daemon is running."""
        installed, _, _ = self._run_command(["docker", "--version"])
        running, _, _ = self._run_command(["docker", "info"], check_running=True)
        self.checks["docker_installed"] = installed
        self.checks["docker_running"] = running
        return installed and running

    def run_all_checks(self):
        """Runs all checks and prints a summary."""
        self.check_docker()

        print("Environment Check Summary:")
        for check, result in self.checks.items():
            if "version" in check:
                print(f"- {check.replace('_',' ').title()}: {result}")
            else:
                status = "OK" if result else "FAILED"
                print(f"- {check.replace('_',' ').title()}: {status}")

        return all(self.checks.values())



import docker
import tarfile
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor

from transitions import Machine
from transitions.extensions.states import Timeout, add_state_features

from convisoappsec.common.docker import SCSCommon
from convisoappsec.logger import LOGGER

RAW_STATE_MSG = 'Scanner {} entered on {} state'


class SARIFParsingError(BaseException):
    pass


class PropertyRequiredError(SARIFParsingError):
    def __init__(self, stderr_log=''):
        pretty_error = self.__parse_pretty_property_error(stderr_log)
        print('Error:', pretty_error)

    def __parse_pretty_property_error(self, stderr_logs):
        expected_error_line = ''

        for log_line in stderr_logs.split('\n'):
            expected_error_text = 'PropertyRequiredError'
            if expected_error_text in log_line:
                expected_error_line = log_line
                break

        error = self.__extract_text_after_colon(expected_error_line)

        return error.strip()

    def __extract_text_after_colon(self, text):
        try:
            return text.split(':', 3)[-1]
        except IndexError:
            return ''


@add_state_features(Timeout)
class ScannerMachine(Machine):
    pass


class ScannerEntity:

    def __init__(self, token, scanner, logger=None, timeout=7200):
        self.logger = logger or LOGGER
        self.token = token

        self.scanner = self.__setup_scanner(scanner)
        self.name = self.scanner.name
        self.results = None

        self.states = [
            'waiting',
            {'name': 'pulling', 'timeout': timeout, 'on_timeout': self._on_timeout},
            {'name': 'running', 'timeout': timeout, 'on_timeout': self._on_timeout},
            {'name': 'sending', 'timeout': timeout, 'on_timeout': self._on_timeout},
            'done'
        ]
        self.machine = ScannerMachine(
            model=self,
            states=self.states,
            initial='waiting'
        )
        self.machine.add_ordered_transitions()
        self._set_callbacks()
        self.to_waiting()

    def __setup_scanner(self, scanner):
        if isinstance(scanner, SCSCommon):
            return scanner
        else:
            return self._instanciate_scanner(scanner)

    def _set_callbacks(self):
        self.machine.on_enter_waiting('_on_waiting')
        self.machine.on_enter_pulling('_on_pulling')
        self.machine.on_enter_running('_on_running')
        self.machine.on_enter_sending('_on_sending')
        self.machine.on_enter_done('_on_done')

    def _instanciate_scanner(self, data):
        return SCSCommon(
            **data,
            token=self.token,
            logger=self.logger,
        )

    def _on_timeout(self):
        self.logger.debug('Scanner {} timeout on state {}'.format(
            self.name, self.state
        ))

    def _on_waiting(self):
        self.logger.debug(RAW_STATE_MSG.format(
            self.name, self.state
        ))

    def _on_pulling(self):
        self.logger.debug(RAW_STATE_MSG.format(
            self.name, self.state
        ))
        image = self.scanner.pull()
        if image:
            self.logger.debug('Image: {}'.format(image))
            self.next_state()
        else:
            raise RuntimeError("Image not found.")

    def _on_running(self):
        self.scanner.run()
        self.end_time = time.time()
        self.logger.debug('Total execution time for {} was {:2f}'.format(
            self.scanner.repository_name,
            self.end_time - self.start_time
        ))
        self.next_state()

    def _on_sending(self):
        self.logger.debug(RAW_STATE_MSG.format(
            self.name, self.state
        ))
        self.results = self.scanner.get_container_reports()
        self.next_state()

    def _on_done(self):
        self.logger.debug(RAW_STATE_MSG.format(
            self.scanner.repository_name, self.state
        ))
        self.scanner.container.remove(v=True, force=True)

    def start(self):
        self.start_time = time.time()
        self.to_pulling()


class ContainerWrapper:

    def __init__(self, token, containers_map, logger, timeout, max_workers=5):
        self.token = token
        self.logger = logger or LOGGER
        self.max_workers = max_workers
        self.scanners = [
            ScannerEntity(
                token=token,
                scanner=scanner,
                logger=logger,
                timeout=timeout
            )
            for scanner in containers_map.values()
        ]

    def run(self):
        self.logger.debug("Starting Execution")
        with ThreadPoolExecutor(max_workers=self.max_workers) as exeggutor:
            for scanner in self.scanners:
                exeggutor.submit(scanner.start)


def convert_sarif_to_sastbox1(report_filepath, repository_dir, container_registry_token, scanner_timeout=7200):
    """
    Args:
        report_filepath (str): filepath to the report to be converted
        repository_dir (str): filepath to the repository being tested
        token (str): Conviso container registry token
        scanner_timeout (int): container timeout

    Returns:
        string: filepath to the converted report
    """
    CONTAINER_IMAGE_NAME = 'sastbox-converter-tool'
    CONTAINER_IMAGE_TAG = 'cc50dee'

    CONTAINER_INPUT_FILEPATH = '/code{}'.format(
        report_filepath.replace(repository_dir, '')
    )
    CONTAINER_OUTPUT_FILENAME = CONTAINER_INPUT_FILEPATH.replace(
        'sarif', 'json'
    )

    CONTAINERS_MAP = {
        CONTAINER_IMAGE_NAME: {
            'repository_dir': repository_dir,
            'repository_name': CONTAINER_IMAGE_NAME,
            'tag': CONTAINER_IMAGE_TAG, 
            'command': [
                '--format', 'sastbox1',
                '--input', CONTAINER_INPUT_FILEPATH,
                '--output', CONTAINER_OUTPUT_FILENAME
            ],
        },
    }
    converter_wrapped = ContainerWrapper(
        token=container_registry_token,
        containers_map=CONTAINERS_MAP,
        logger=None,
        timeout=scanner_timeout
    )

    converter_wrapped.logger.setLevel('WARN')
    converter_wrapped.run()
    converter_wrapped.logger.setLevel('INFO')

    scanner = converter_wrapped.scanners[0].scanner
    last_scan_name = scanner.name
    last_container = scanner.docker.containers.get(
        last_scan_name
    )

    try:
        chunks, _ = last_container.get_archive(CONTAINER_OUTPUT_FILENAME)
        output_filepath = __extract_tarball_chunks(
            chunks, report_filepath.replace('sarif', 'json')
        )
    except docker.errors.APIError as error:
        stderr_log = last_container.logs(stderr=True).decode('utf-8')
        raise PropertyRequiredError(stderr_log)

    return output_filepath


def __extract_tarball_chunks(tarball_chunks, report_absolute_filepath):
    """

    Args:
        tarball_chunks (int): The number of bytes returned by each iteration of the generator
        report_filename (string): The name of the extracted report

    Returns:
        string: Report absolute filepath in local filesystem
    """
    output_dirpath = report_absolute_filepath[
        :report_absolute_filepath.rfind('/')
    ]

    with tempfile.TemporaryFile() as tmp_wrapper_file:
        for chunk in tarball_chunks:
            tmp_wrapper_file.write(chunk)
        tmp_wrapper_file.seek(0)

        with tarfile.open(mode="r|", fileobj=tmp_wrapper_file) as talball_file:
            talball_file.extractall(path=output_dirpath)

    return report_absolute_filepath



from json import dumps as json_dumps

import click
import click_log

from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import on_http_error
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.logger import LOGGER

click_log.basic_config(LOGGER)


@click.command()
@click_log.simple_verbosity_option(LOGGER)
@click.option(
    '-c',
    '--company-id',
    type=int,
    required=True,
    help="The company ID to have its resources used.",
)
@click.option(
    '-p',
    '--page',
    type=int,
    default=1,
    show_default=True,
    required=False,
    help="Page to be consulted",
)
@click.option(
    '-l',
    '--limit',
    type=int,
    default=32,
    show_default=True,
    required=False,
    help="Items limit per page.",
)
@help_option
@pass_flow_context
def ls(flow_context, company_id, asset_name="", page=1, limit=32):
    try:
        conviso_api = flow_context.create_conviso_graphql_client()

        perform_command(conviso_api, company_id, asset_name, page, limit)

    except Exception as exception:
        on_http_error(exception)
        raise click.ClickException(str(exception)) from exception


def perform_command(conviso_api, company_id, asset_name="", page=1, limit=32):
    assets_found = conviso_api.assets.get_by_company_id_or_name(
        company_id,
        asset_name,
        page,
        limit
    )

    indented_output = json_dumps(assets_found, indent=2)
    print(indented_output)



import tarfile
import tempfile
from contextlib import suppress
from datetime import datetime
from threading import Lock, Thread
from uuid import uuid4

from convisoappsec.flow.util import SourceCodeCompressor
from convisoappsec.logger import LOGGER

import docker

from .exceptions import *

mutex = Lock()


class EventLogging(Thread):

    def __init__(self, logger, events):
        super().__init__(name='Docker-Events', daemon=True)
        self.logger = logger or LOGGER
        self.events = events

    def docker_log(self, event):
        ts = datetime.fromtimestamp(event['time']).strftime('%d/%m/%Y %H:%M:%S')
        message = "{timestamp}|{Type} {Action} on {Actor[ID]}".format(**event, timestamp=ts)
        self.logger.debug(message)

    def run(self):
        for event in self.events:
            if 'conviso-cli-' in event['Actor']['ID']:
                self.docker_log(event)
        self.logger.debug('Docker Events closed')


class Credentials:

    AUTHS = {}

    def __init__(self, docker, logger):
        self.docker = docker
        self.logger = logger or LOGGER

    def login(self, registry, password, username='AWS'):
        mutex.acquire()
        if not registry in Credentials.AUTHS.keys():
            self.logger.debug(
                "   \U0001F511 Checking Authorization for {0}".format(registry))
            login_args = {
                'registry': registry,
                'username': username,
                'password': password,
                'reauth': True,
            }
            try:
                login_result = self.docker.login(**login_args)
                status = login_result['Status']
                self.logger.debug("Login result was {0}".format(login_result))
                if status == 'Login Succeeded':
                    Credentials.AUTHS[registry] = password
                    self.logger.debug('   \U0001F513 ' + status)
            except docker.errors.APIError:
                self.logger.debug("Docker API Error")
        mutex.release()


class SCSCommon:

    # General Settings
    DEFAULT_REGISTRY = 'public.ecr.aws/convisoappsec'
    DEFAULT_TAG = 'latest'
    DEFAULT_CONTAINER_CODE_DIR = '/code'
    SUCCESS_EXIT_CODE = 0
    DOCKER_CLIENT = None
    DOCKER_CREDENTIALS = None
    DOCKER_EVENTS = None
    DOCKER_EVENTS_THREAD = None
    TEMPDIR = None

    def __init__(self, tag=None, registry=None, repository_name=None, token=None, logger=None, command=None, repository_dir=None):
        self.logger = logger or LOGGER
        self.token = token
        uuid = str(uuid4())
        self.docker = self._get_docker_client()
        self.__container_name = "conviso-cli-{0}".format(uuid)
        self.__docker_events = self._get_docker_events()
        self.__docker_events_thread = self._get_docker_events_thread()
        self.__credentials = self._get_docker_credentials()
        self.__source_code_volume_name = "conviso-cli-{0}".format(uuid)

        self.container = None
        self.tag = tag or self.DEFAULT_TAG
        self.registry = registry or self.DEFAULT_REGISTRY
        self.repository_name = repository_name
        self.command = command
        self.repository_dir = repository_dir
        if token:
            self._login(self.token)

    def _get_docker_client(self):
        mutex.acquire()
        if not SCSCommon.DOCKER_CLIENT:
            SCSCommon.DOCKER_CLIENT = docker.from_env(
                version="auto"
            )
        mutex.release()
        return SCSCommon.DOCKER_CLIENT

    def _get_docker_credentials(self):
        mutex.acquire()
        if not SCSCommon.DOCKER_CREDENTIALS:
            SCSCommon.DOCKER_CREDENTIALS = Credentials(
                docker=self.docker,
                logger=self.logger
            )
        mutex.release()
        return SCSCommon.DOCKER_CREDENTIALS

    def _get_docker_events(self):
        mutex.acquire()
        if not SCSCommon.DOCKER_EVENTS:
            SCSCommon.DOCKER_EVENTS = self.docker.events(decode=True)
        mutex.release()
        return SCSCommon.DOCKER_EVENTS

    def _get_docker_events_thread(self):
        mutex.acquire()
        if not SCSCommon.DOCKER_EVENTS_THREAD:
            SCSCommon.DOCKER_EVENTS_THREAD = EventLogging(
                logger=self.logger,
                events=self.__docker_events,
            )
            SCSCommon.DOCKER_EVENTS_THREAD.start()
        mutex.release()
        return SCSCommon.DOCKER_EVENTS_THREAD

    def _login(self, password, username='AWS'):
        self.__credentials.login(
            registry=self.registry,
            password=password,
        )

    def _create_host_output_directory(self):
        """Creates a directory on the host machine to store container outputs.

        Returns:
            string: absolute path to the directory
        """
        if not SCSCommon.TEMPDIR:
            SCSCommon.TEMPDIR = tempfile.mkdtemp(
                prefix='conviso-output-',
                dir=self.repository_dir
            )

        self.logger.debug("Created output directory at {}".format(
            SCSCommon.TEMPDIR
        ))
        return SCSCommon.TEMPDIR

    @property
    def size(self):
        registry_data = self.docker.images.get_registry_data(self.image)

        descriptor = registry_data.attrs.get('Descriptor', {})
        return descriptor.get('size') * 1024 * 1024

    def run(self):
        container = self.__create_container()
        self.__inject_container_source_code(container)
        container.start()

        container_stderr = container.logs(
            stream=True,
            stdout=False,
            stderr=True
        )
        for message in container_stderr:
            self.logger.debug(message)

        container_stdout = container.logs(
            stream=True,
            stdout=True,
            stderr=False,
        )

        for message in container_stdout:
            self.logger.debug(message)

    def wait(self):
        wait_result = self.__container.wait()
        status_code = wait_result.get('StatusCode')

        if not status_code == self.SUCCESS_EXIT_CODE:
            raise CommonException()

        return status_code

    def __has_method(self, method_name):
        return hasattr(self, method_name)

    def pull(self):
        '''
        Acts as docker pull <image_name>, downloading the image configured in the instance.

        Returns:
            # Example:
            {
                'status': 'Downloading',
                'progressDetail': {'current': int, 'total': int},
                'id': 'string'
            }
        '''
        self.logger.debug(
            "Pulling scanner image {}. It might takes some minutes.".format(
                self.repository_name)
        )
        if self.has_pre_pull:
            self._pre_pull()
            self.logger.debug("End of pre_pull method.")
        try:
            return self.docker.images.pull(
                repository=self.repository,
                tag=self.tag
            )
        except docker.errors.ImageNotFound:
            self.logger.debug(
                "Image {} not found on registry".format(self.image)
            )
            return False

    def __get_container(self):
        return self.docker.containers.get(
            self.__container_name
        )

    @property
    def __container(self):
        try:
            return self.__get_container()
        except:
            return self.__create_container()

    def __create_container(self):
        self.logger.debug(
            "Creating Container {0}".format(self.__container_name)
        )
        return self.docker.containers.create(
            self.image,
            name=self.__container_name,
            volumes=self.volumes,
            detach=True,
            command=self.command
        )

    @property
    def repository(self):
        return "{registry}/{repository_name}".format(
            registry=self.registry,
            repository_name=self.repository_name,
        )

    @property
    def image(self):
        return "{repository}:{tag}".format(
            repository=self.repository,
            tag=self.tag,
        )

    @property
    def has_pre_pull(self):
        return self.__has_method('_pre_pull')

    @property
    def has_read_scan_stdout(self):
        return self.__has_method('_read_scan_stdout')

    @property
    def has_read_scan_stderr(self):
        return self.__has_method('_read_scan_stderr')

    @property
    def volumes(self):
        return {
            self.__source_code_volume_name: {
                'bind': SCSCommon.DEFAULT_CONTAINER_CODE_DIR,
                'mode': 'rw',
            }
        }

    @property
    def name(self):
        return self.__container_name

    def __del__(self):
        self.logger.debug(
            "Removing container {0}".format(self.__container_name)
        )
        with suppress(Exception):
            self.container.remove(v=True, force=True)

    def __inject_container_source_code(self, container):
        try:
            self.logger.debug(
                "Loading Source code into {0}".format(self.__container_name)
            )

            with tempfile.TemporaryFile() as wrapper_file:
                compressor = SourceCodeCompressor(
                    self.repository_dir
                )
                compressor.write_to(wrapper_file)
                wrapper_file.seek(0)
                container.put_archive(
                    SCSCommon.DEFAULT_CONTAINER_CODE_DIR,
                    wrapper_file
                )

        except Exception as e:
            self.logger.error(e)

    def _get_container_tarball_chunks(self, container_filepath):
        """From the path provided you get the file binary/chunks from the container.

        Args:
            container_filepath (string): Absolute Path of the file within the container

        Returns:
            int: For further information see https://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.Container.get_archive
        """
        try:
            container = self.__container
            chunks, _ = container.get_archive(container_filepath)
            return chunks

        except docker.errors.NotFound:
            self.logger.debug("{} does not detected issues, continuing...".format(
                self.repository_name
            ))
            return None

    def _extract_tarball_chunks(self, tarball_chunks, report_filename):
        """

        Args:
            tarball_chunks (int): The number of bytes returned by each iteration of the generator
            report_filename (string): The name of the extracted report

        Returns:
            string: Report absolute filepath in local filesystem
        """
        output_dirpath = self._create_host_output_directory()
        report_absolute_filepath = '{}/{}'.format(
            output_dirpath, report_filename
        )

        with tempfile.TemporaryFile() as tmp_wrapper_file:
            for chunk in tarball_chunks:
                tmp_wrapper_file.write(chunk)
            tmp_wrapper_file.seek(0)

            with tarfile.open(mode="r|", fileobj=tmp_wrapper_file) as talball_file:
                talball_file.extractall(path=output_dirpath)

        self.logger.debug("report from {} saved to {}".format(
            self.__container_name,
            report_absolute_filepath
        ))
        return report_absolute_filepath

    def get_container_reports(self):
        """ Get all reports found in the instantiated container and save on the host machine.

        Returns:
            string: Report absolute filepath in local filesystem
        """
        self.logger.debug(
            "Collecting reports from {0}".format(self.__container_name)
        )
        try:
            possible_report_extensions = ['sarif', 'json']
            for ext in possible_report_extensions:
                report_filename = '{}.{}'.format(self.repository_name, ext)

                container_filepath = '/{}'.format(report_filename)
                tarball_chunks = self._get_container_tarball_chunks(
                    container_filepath
                )
                if not tarball_chunks:
                    continue

                absolute_report_path = self._extract_tarball_chunks(
                    tarball_chunks,
                    report_filename
                )
                return absolute_report_path

        except Exception as e:
            self.logger.error(e)



class IssuesInput:
    def __init__(self, company_id, asset_id, page=1, per_page=50):
        self.company_id = company_id
        self.asset_id = asset_id
        self.page = page
        self.per_page = per_page
        self.statuses = ['CREATED', 'IDENTIFIED']

    def to_graphql_dict(self):
        return {
            "companyId": self.company_id,
            "page": self.page,
            "perPage": self.per_page,
            "assetIds": self.asset_id,
            "statuses": self.statuses
        }



import json
import click
import click_log
import traceback
from convisoappsec.common.retry_handler import RetryHandler
from copy import deepcopy as clone
from convisoappsec.common.box import ContainerWrapper
from convisoappsec.flow.graphql_api.beta.models.issues.iac import CreateIacFindingInput
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import asset_id_option, on_http_error
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.logger import LOGGER, log_and_notify_ast_event
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from convisoappsec.flow import GitAdapter
from convisoappsec.common.graphql.errors import ResponseError
from convisoappsec.common.cleaner import Cleaner

click_log.basic_config(LOGGER)


@click.command()
@click_log.simple_verbosity_option(LOGGER)
@asset_id_option(required=False)
@click.option(
    '-r',
    '--repository-dir',
    default=".",
    show_default=True,
    type=click.Path(
        exists=True,
        resolve_path=True,
    ),
    required=False,
    help="The source code repository directory.",
)
@click.option(
    "--send-to-flow/--no-send-to-flow",
    default=True,
    show_default=True,
    required=False,
    help="""Enable or disable the ability of send analysis result
    reports to flow.""",
    hidden=True
)
@click.option(
    "--scanner-timeout",
    hidden=True,
    required=False,
    default=7200,
    type=int,
    help="Set timeout for each scanner"
)
@click.option(
    "--parallel-workers",
    hidden=True,
    required=False,
    default=2,
    type=int,
    help="Set max parallel workers"
)
@click.option(
    "--deploy-id",
    default=None,
    required=False,
    hidden=True,
    envvar=("CONVISO_DEPLOY_ID", "FLOW_DEPLOY_ID")
)
@click.option(
    '--experimental',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable experimental features.",
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@click.option(
    '--from-ast',
    default=False,
    is_flag=True,
    hidden=True,
    help="Internal use only.",
)
@click.option(
    '--cleanup',
    default=False,
    is_flag=True,
    show_default=True,
    help="Clean up system resources, including temporary files, stopped containers, unused Docker images and volumes.",
)
@click.option(
    '--control-sync-status-id',
    required=False,
    hidden=True,
    help="Control sync status id.",
)
@click.option(
    '--control-sync-status-id',
    required=False,
    hidden=True,
    help="Control sync status id.",
)
@help_option
@pass_flow_context
@click.pass_context
def run(context, flow_context, asset_id, company_id, repository_dir, send_to_flow, scanner_timeout,
        parallel_workers, deploy_id, experimental, asset_name, from_ast, cleanup, control_sync_status_id):
    """
      This command will perform IAC analysis at the source code. The analysis
      results can be reported or not to flow application.
    """
    if not from_ast:
        prepared_context = RequirementsVerifier.prepare_context(clone(context))

        params_to_copy = [
            'asset_id', 'company_id', 'repository_dir', 'send_to_flow',
            'deploy_id', 'scanner_timeout', 'parallel_workers', 'experimental', 'cleanup'
        ]

        for param_name in params_to_copy:
            context.params[param_name] = (
                    locals()[param_name] or prepared_context.params[param_name]
            )

    perform_command(
        flow_context, context.params['asset_id'], context.params['company_id'], context.params['repository_dir'],
        context.params['send_to_flow'], context.params['scanner_timeout'], context.params['deploy_id'],
        context.params['experimental'], context.params['cleanup'], from_ast, control_sync_status_id
    )


def deploy_results_to_conviso(
        conviso_api, results_filepaths, asset_id, company_id, flow_context, deploy_id, commit_ref=None, control_sync_status_id=None
):
    results_context = click.progressbar(results_filepaths, label="Sending results to the Conviso Platform...")

    with results_context as reports:
        for report_path in reports:
            try:
                with open(report_path) as report_file:
                    data = parse_data(json.load(report_file))
            except Exception:
                LOGGER.warn(f"⚠️ Error processing report file. Our technical team has been notified.")
                full_trace = traceback.format_exc()
                log_and_notify_ast_event(
                    flow_context=flow_context, company_id=company_id, asset_id=asset_id,
                    ast_log=str(full_trace)
                )
                continue

            for issue in data:
                try:
                    issue_model = CreateIacFindingInput(
                        asset_id=asset_id,
                        file_name=issue.get("file_name"),
                        vulnerable_line=issue.get("vulnerable_line"),
                        title=issue.get("title"),
                        description=issue.get("description"),
                        severity=issue.get("severity"),
                        code_snippet=parse_code_snippet(issue.get("code_snippet")),
                        reference=parse_conviso_references(issue.get("reference", "")),
                        first_line=issue.get("first_line"),
                        category=format_cwe_id(issue.get("cwe")),
                        original_issue_id_from_tool=issue.get('hash_issue', []),
                        solution=issue.get("solution"),
                        control_sync_status_id=control_sync_status_id
                    )

                    conviso_api.issues.create_iac(issue_model)

                except ResponseError as error:
                    if error.code == 'RECORD_NOT_UNIQUE':
                        continue
                    else:
                        retry_handler = RetryHandler(
                            flow_context=flow_context, company_id=company_id, asset_id=asset_id
                        )
                        retry_handler.execute_with_retry(conviso_api.issues.create_iac, issue_model)
                except Exception:
                    retry_handler = RetryHandler(
                        flow_context=flow_context, company_id=company_id, asset_id=asset_id
                    )
                    retry_handler.execute_with_retry(conviso_api.issues.create_iac, issue_model)

                    continue


def perform_command(
        flow_context, asset_id, company_id, repository_dir, send_to_flow, scanner_timeout,
        deploy_id, experimental, cleanup, from_ast, control_sync_status_id
):

    if send_to_flow and experimental and not asset_id:
        raise click.MissingParameter(
            "It is required when sending reports to Conviso Platform using experimental API.",
            param_type="option",
            param_hint="--asset-id",
        )

    try:
        REQUIRED_CODEBASE_PATH = '/code'
        IAC_IMAGE_NAME = 'iac_scanner_checkov'
        IAC_SCAN_FILENAME = '/{}.json'.format(IAC_IMAGE_NAME)
        containers_map = {
            IAC_IMAGE_NAME: {
                'repository_dir': repository_dir,
                'repository_name': IAC_IMAGE_NAME,
                'tag': 'unstable',
                'command': [
                    '-c', REQUIRED_CODEBASE_PATH,
                    '-o', IAC_SCAN_FILENAME,
                ],
            },
        }

        LOGGER.info('💬 Preparing Environment...')
        conviso_rest_api = flow_context.create_conviso_rest_api_client()
        token = conviso_rest_api.docker_registry.get_sast_token()
        scanners_wrapper = ContainerWrapper(
            token=token,
            containers_map=containers_map,
            logger=LOGGER,
            timeout=scanner_timeout
        )

        LOGGER.info('💬 Starting IaC...')
        scanners_wrapper.run()

        results_filepaths = []
        for r in scanners_wrapper.scanners:
            report_filepath = r.results
            if report_filepath:
                results_filepaths.append(report_filepath)

        LOGGER.info('💬 Processing Results...')
        if send_to_flow:
            git_adapater = GitAdapter(repository_dir)
            end_commit = git_adapater.head_commit
            conviso_beta_api = flow_context.create_conviso_api_client_beta()

            deploy_results_to_conviso(
                conviso_beta_api, results_filepaths, asset_id, company_id, flow_context, deploy_id=deploy_id,
                commit_ref=end_commit, control_sync_status_id=control_sync_status_id
            )
        LOGGER.info('✅ IaC Scan Finished.')

        if cleanup and from_ast == False:
            LOGGER.info("🧹 Cleaning up ...")
            cleaner = Cleaner()
            cleaner.cleanup()

    except Exception as e:
        on_http_error(e)
        raise click.ClickException(str(e)) from e


def parse_conviso_references(references=[]):
    DIVIDER = "\n"

    return DIVIDER.join(references)


def parse_code_snippet(code_snippet):
    lines = code_snippet.split("\n")
    cleaned_lines = [line.rstrip() for line in lines if line.strip()]
    code_snippet = "\n".join(cleaned_lines)

    return code_snippet


def format_cwe_id(cwe_input):
    cwe_str = str(cwe_input).strip()

    if cwe_str.upper().startswith("CWE-"):
        return cwe_str
    return f"CWE-{cwe_str}"


def parse_data(sarif_result):
    vulnerabilities = []

    for result in sarif_result['runs'][0]['results']:
        
        vulnerability = {
            'file_name': result['locations'][0]['physicalLocation']['artifactLocation']['uri'],
            'vulnerable_line': result['locations'][0]['physicalLocation']['region']['startLine'],
            'code_snippet': result['locations'][0]['physicalLocation']['contextRegion']['snippet']['text'],
            'title': sarif_result['runs'][0]['tool']['driver']['rules'][result['ruleIndex']]['name'],
            'description': result['message']['text'],
            'severity': result['level'],
            'first_line': result['locations'][0]['physicalLocation']['contextRegion']['startLine'],
            'cwe': result['properties']['cweId'],
            'solution': result['properties']['solution'],
            'hash_issue': result['partialFingerprints']['hashIssueV2']
        }
        vulnerabilities.append(vulnerability)

    return vulnerabilities


EPILOG = '''
Examples:

  \b
  1 - Reporting the results to Conviso Platform API:
    1.1 - Running an analysis at all commit range:
      $ export CONVISO_API_KEY='your-api-key'
      $ {command}

'''  # noqa: E501

SHORT_HELP = "Perform Infrastructure Code analysis"

command = 'conviso iac run'
run.short_help = SHORT_HELP
run.epilog = EPILOG.format(
    command=command,
)



import os
import shutil
import tempfile
import docker

class Cleaner:
    """Responsible for cleaning temporary files, Docker containers, images, and volumes."""

    def __init__(self):
        try:
            self.client = docker.from_env()
        except Exception as e:
            print(f"Error initializing Docker client: {e}")
            self.client = None

    def cleanup(self):
        """ Responsable to clean dirs, docker images and containers after all executions,
            removes all stopped containers, unused networks, dangling images, and build cache.
        """
        try:
            client = docker.from_env()
            self.perform_cleanup()

            for container in client.containers.list(all=True):
                try:
                    container.remove()
                except Exception:
                    continue

            for image in client.images.list():
                if image.tags and any(tag.startswith("public.ecr.aws/convisoappsec/") for tag in image.tags):
                    try:
                        client.images.remove(image.id)
                    except Exception as e:
                        print(f"Error removing image {image.tags}: {e}")
                        continue

            volumes = client.volumes.list()
            for volume in volumes:
                try:
                    volume.remove()
                except Exception:
                    continue

        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            return

    def perform_cleanup(self):
       """Method to clean the tmp directory and remove 'conviso-output-' directories in the current directory."""

       tmp_dir = tempfile.gettempdir()

       # Clear system temp directory
       try:
           for filename in os.listdir(tmp_dir):
               file_path = os.path.join(tmp_dir, filename)
               try:
                   if os.path.isfile(file_path) or os.path.islink(file_path):
                       os.remove(file_path)
                   elif os.path.isdir(file_path):
                       shutil.rmtree(file_path)
               except Exception:
                   pass
       except Exception:
           pass

       # Clear 'conviso-output-' directories in the current directory
       try:
           for filename in os.listdir("."):
               dir_path = os.path.join(".", filename)
               if os.path.isdir(dir_path) and filename.startswith("conviso-output-"):
                   try:
                       shutil.rmtree(dir_path)
                   except Exception:
                       pass
       except Exception:
           pass




from git import Repo
from git.exc import GitCommandError, InvalidGitRepositoryError
from giturlparse import parse as gitparse
from hashlib import sha256
import os

from convisoappsec.flow.version_control_system_adapter import GitAdapter
from convisoappsec.logger import LOGGER


class DescriptionParsingError(GitCommandError):
    pass


class GitDataParser:
    def __init__(self, repository_dir):
        try:
            self.__validate_git_repo(repository_dir)

            self.repository_dir = repository_dir
            self._git_adapter = GitAdapter(repository_dir)

        except InvalidGitRepositoryError as exp:
            LOGGER.error(
                'Invalid Git repository: "{}".'.format(repository_dir)
            )
            raise exp

        except Exception as exp:
            LOGGER.error("An unxpected error occurred.")
            raise exp

    def parse_name(self):
        try:
            git_remote_url = self._git_adapter._git_client.ls_remote(
                '--get-url'
            )

            return self.__parse_name_from_git_url(git_remote_url)
        except GitCommandError:
            dirname = os.path.dirname(self.repository_dir)
            basename = os.path.basename(self.repository_dir)
            digest = sha256(dirname.encode('utf-8')).hexdigest()

            return basename + '-' + digest
        except Exception as exception:
            raise exception

    def __parse_description(self):
        try:
            readme = self._git_adapter._git_client.show(':README.md')
            readme = readme.replace('\n', '\n<br>')

            return readme

        except DescriptionParsingError as exception:
            raise exception

        except Exception as exception:
            raise exception

    def __parse_name_from_git_url(self, git_url):
        parser = gitparse(git_url)

        return '{} ({}/{})'.format(
            parser.name,
            parser.resource,
            parser.owner,
        )

    def __validate_git_repo(self, repository_dir):
        repo = Repo(repository_dir)
        repo.git.config("--global", "--add", "safe.directory", repository_dir)

        return repo



import click

from convisoappsec.flowcli import help_option
from .generate import generate


@click.group()
@help_option
def sbom():
    pass


sbom.add_command(generate)

sbom.epilog = '''
  Run conviso sbom COMMAND --help for more information on a command.
'''



from convisoappsec.flow.source_code_scanner import SCC
from convisoappsec.logger import LOGGER
import docker

def project_metrics(source_code_dir):
    try:
        scanner = SCC(source_code_dir, create_source_code_volume=False)
        scanner.scan()
        return {
            'total_lines': scanner.total_source_code_lines
        }
    except docker.errors.APIError as e:
        LOGGER.error('Error on fetch project metrics')
        LOGGER.exception(e)
        return {}




import click

from convisoappsec.flowcli import help_option
from .run import run


@click.group()
@help_option
def sast():
    pass


sast.add_command(run)

sast.epilog = '''
  Run flow sast COMMAND --help for more information on a command.
'''



import click

from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import DeployFormatter

from .context import pass_create_context
from .with_ import with_


@click.group()
@click.option(
    '-f',
    '--output-format',
    required=False,
    type=click.Choice(DeployFormatter.FORMATS()),
    default=DeployFormatter.DEFAULT,
    show_default=True,
)
@help_option
@pass_create_context
def create(create_context, output_format):
    create_context.output_formatter = DeployFormatter(
        format=output_format,
    )


create.add_command(with_)

create.epilog = '''
  Run flow deploy create COMMAND --help for more information on a command.
'''



class AssetInput:
    def __init__(self, company_id, name, scan_type=None, repo_url=None):
        self.name = name
        self.company_id = company_id
        self.scan_type = scan_type
        self.repo_url = repo_url

    def to_graphql_dict(self):
        return {
            "companyId": self.company_id,
            "name": self.name,
            "scanType": self.scan_type,
            "repoUrl": self.repo_url
        }



class Normalize:

    @staticmethod
    def normalize_severity(severity):
        """
        The function normalizes severity by validating and returning a standardized severity level.
        """

        validate_severity = ["LOW", "MEDIUM", "HIGH", "CRITICAL", "NOTIFICATION"]
        if severity.upper() in validate_severity:
            return severity.upper()
        else:
            return validate_severity[0]



import click

from convisoappsec.flowcli import help_option

from .context import pass_tag_tracker_context
from .sort_by import sort_by


@click.group()
@help_option
@click.option(
    "-r",
    "--repository-dir",
    required=False,
    type=click.Path(exists=True, resolve_path=True),
    default='.',
    show_default=True,
    help="Repository dir to track version tags.",
)
@pass_tag_tracker_context
def tag_tracker(tag_tracker_context, repository_dir):
    tag_tracker_context.repository_dir = repository_dir


tag_tracker.add_command(sort_by)

tag_tracker.epilog = '''
  Run flow deploy create with tag-tracker COMMAND --help for
  more information on a command.
'''



import click

from convisoappsec.common import safe_join_url
from convisoappsec.flow import api
from convisoappsec.flow.graphql_api.beta.client import ConvisoGraphQLClientBeta
from convisoappsec.flow.graphql_api.v1.client import ConvisoGraphQLClient
from convisoappsec.version import __version__


class FlowContext(object):
    def __init__(self):
        self.key = None
        self.url = None
        self.insecure = None
        self.ci_provider = None
        self.logger = None

    def create_conviso_rest_api_client(self):
        return api.RESTClient(
            key=self.key,
            url=self.url,
            insecure=self.insecure,
            user_agent={
                'name': 'flowcli',
                'version': __version__,
            },
            ci_provider_name=self.ci_provider.name
        )

    def create_conviso_graphql_client(self):
        url = safe_join_url(self.url, "/graphql")

        return ConvisoGraphQLClient(
            api_url=url,
            api_key=self.key
        )

    def create_conviso_api_client_beta(self):
        url = safe_join_url(self.url, "/graphql")

        return ConvisoGraphQLClientBeta(
            api_url=url,
            api_key=self.key,
        )


pass_flow_context = click.make_pass_decorator(
    FlowContext, ensure=True
)



import click
from convisoappsec.flowcli import help_option
from .assert_security_rules import assert_security_rules
from .run import run


@click.group()
@help_option
def vulnerability():
    pass


vulnerability.add_command(assert_security_rules)
vulnerability.add_command(run)

vulnerability.epilog = '''
  Run flow vulnerability COMMAND --help for more information on a command.
'''



import click
from pathlib import Path

from convisoappsec.flow.version_control_system_adapter import GitAdapter
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.flowcli import help_option


def finding_report_files_validation(_, __, finding_report_files):
    allowed_extensions = ['.json']
    allowed_extensions_str = '[{0}]'.format(
        '|'.join(allowed_extensions)
    )

    for report_file in finding_report_files:
        file_name = report_file.name

        if Path(file_name).suffix not in allowed_extensions:
            raise click.BadParameter(
                'Allowed file extensions {0}. File given: {1}'.format(
                    allowed_extensions_str,
                    file_name
                )
            )

    return finding_report_files


@click.command()
@click.argument(
    "finding-report-files",
    nargs=-1,
    required=False,
    type=click.File('r'),
    callback=finding_report_files_validation
)
@click.option(
    '-c',
    '--current-commit',
    required=False,
    help="""The hash of current commit. If no value is set
    so the HEAD commit will be used.""",
)
@click.option(
    "-r",
    "--repository-dir",
    required=False,
    type=click.Path(exists=True),
    default=".",
    show_default=True,
    help='The root dir of repository.',
)
@click.option(
    "-t",
    "--default-report-type",
    required=False,
    help="""This value is used when report type attribute is missing.
    This not override an existing value at report file""",
)
@help_option
@pass_flow_context
def version_tracker(
    flow_context, finding_report_files,
    repository_dir, current_commit, default_report_type
):
    try:
        if not finding_report_files:
            click.echo(
                "Nothing to be done, finding report files argument empty"
            )

            return

        git = GitAdapter(repository_dir)
        current_commit = current_commit or git.head_commit
        commit_refs = git.show_commit_refs(current_commit)
        finding_report_files_prog_bar_ctx = click.progressbar(
            finding_report_files,
            label='Sending finding reports'
        )

        flow = flow_context.create_conviso_rest_api_client()

        with finding_report_files_prog_bar_ctx as reports_with_progressbar:
            for finding_report in reports_with_progressbar:
                flow.findings.create(
                    commit_refs,
                    finding_report,
                    default_report_type=default_report_type
                )

    except Exception as e:
        raise click.ClickException(str(e)) from e
